#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fermionic instructions
----------------------

Extra instructions specific for Fermionic quantum computing.
"""

from piquasso.api.instruction import Preparation, Gate


class ParentHamiltonian(Preparation):
    r"""Prepares the fermionic Gaussian state with a specified parent Hamiltonian.

    The density matrix is of the form

    .. math::

        \rho = \frac{e^{\hat{H}}}{\operatorname{Tr} e^{\hat{H}}},

    where :math:`\hat{H}` is the parent Hamiltonian and has the form

    .. math::

        \hat{H} &= \mathbf{f}^\dagger H \mathbf{f}, \\\\
        H &= \begin{bmatrix}
            - \overline{A} & B \\
            - \overline{B} & A
        \end{bmatrix}. \\\\

    Here, :math:`A` is self-adjoint and :math:`B` is skew-symmetric.
    """

    def __init__(self, hamiltonian):
        super().__init__(params=dict(hamiltonian=hamiltonian))


class GaussianHamiltonian(Gate):
    r"""Applies a fermionic Gaussian transformation using its quadratic Hamiltonian.

    The gate unitary is of the form

    .. math::

        U = e^{i \hat{H}},

    where

    .. math::
        \hat{H} &= \mathbf{f} H \mathbf{f}^\dagger, \\\\
        H &= \begin{bmatrix}
            A & -\overline{B} \\
            B & -\overline{A}
        \end{bmatrix}.

    Here, :math:`A` is self-adjoint and :math:`B` is skew-symmetric.
    """

    def __init__(self, hamiltonian):
        super().__init__(params=dict(hamiltonian=hamiltonian))


class ControlledPhase(Gate):
    r"""Controlled-phase gate.

    The controlled-phase gate is defined via the unitary operator

    .. math::
        \text{CPhase}_{ij} (\phi) = \exp \left ( i \phi n_i n_j \right ).

    Note:
        This is a non-linear gate, therefore it couldn't be used with
        :class:`~piquasso.fermionic.gaussian.simulator.GaussianSimulator`.

    Note:
        This is analoguous to the :class:`~piquasso.instructions.gates.CrossKerr` gate
        in the photonic setting.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, phi: float) -> None:
        """
        Args:
            phi (float): The controlled-phase angle.
        """
        super().__init__(params=dict(phi=phi))


class IsingXX(Gate):
    r"""Ising XX coupling gate.

    The Ising XX gate is defined via the unitary operator

    .. math::
        \text{XX}_{ij} (\phi) = \exp \left ( i \phi X \otimes X \right )
            = \cos \phi I + i \sin \phi (X \otimes X).

    Considering a two-mode system with Majorana operators :math:`m_1, m_2, m_3, m_4`,
    :math:`X \otimes X` can also be written as

    .. math::
        X \otimes X = -i m_2 m_3.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, phi: float) -> None:
        """
        Args:
            phi (float): The rotation angle.
        """
        super().__init__(params=dict(phi=phi))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_program_from_dict_with_StateVector_preparation():
    instructions_dict = {
        "instructions": [
            {
                "type": "StateVector",
                "attributes": {
                    "constructor_kwargs": {
                        "occupation_numbers": [1, 1],
                        "coefficient": 1.0,
                    },
                    "modes": [0, 1],
                },
            },
            {
                "type": "Beamsplitter",
                "attributes": {
                    "constructor_kwargs": {
                        "theta": np.pi / 3,
                        "phi": np.pi / 4,
                    },
                    "modes": [0, 1],
                },
            },
            {
                "type": "ParticleNumberMeasurement",
                "attributes": {
                    "constructor_kwargs": {},
                    "modes": [0],
                },
            },
        ]
    }

    program = pq.Program.from_dict(instructions_dict)

    assert isinstance(program.instructions[0], pq.StateVector)
    assert program.instructions[0].params == {
        "occupation_numbers": (1, 1),
        "coefficient": 1.0,
    }
    assert program.instructions[0].modes == [0, 1]

    assert isinstance(program.instructions[1], pq.Beamsplitter)
    assert program.instructions[1].params == {
        "theta": np.pi / 3,
        "phi": np.pi / 4,
    }
    assert program.instructions[1].modes == [0, 1]

    assert isinstance(program.instructions[2], pq.ParticleNumberMeasurement)
    assert program.instructions[2].params == {}
    assert program.instructions[2].modes == [0]


def test_program_from_dict_with_DensityMatrix_preparation():
    instructions_dict = {
        "instructions": [
            {
                "type": "DensityMatrix",
                "attributes": {
                    "constructor_kwargs": {
                        "bra": [1, 1],
                        "ket": [0, 0],
                        "coefficient": 1.0,
                    },
                    "modes": [0, 1],
                },
            },
            {
                "type": "Beamsplitter",
                "attributes": {
                    "constructor_kwargs": {
                        "theta": np.pi / 3,
                        "phi": np.pi / 4,
                    },
                    "modes": [0, 1],
                },
            },
            {
                "type": "ParticleNumberMeasurement",
                "attributes": {
                    "constructor_kwargs": {},
                    "modes": [0],
                },
            },
        ]
    }

    program = pq.Program.from_dict(instructions_dict)

    assert isinstance(program.instructions[0], pq.DensityMatrix)
    assert program.instructions[0].params == {
        "bra": (1, 1),
        "ket": (0, 0),
        "coefficient": 1.0,
    }
    assert program.instructions[0].modes == [0, 1]

    assert isinstance(program.instructions[1], pq.Beamsplitter)
    assert program.instructions[1].params == {
        "theta": np.pi / 3,
        "phi": np.pi / 4,
    }
    assert program.instructions[1].modes == [0, 1]

    assert isinstance(program.instructions[2], pq.ParticleNumberMeasurement)
    assert program.instructions[2].params == {}
    assert program.instructions[2].modes == [0]


def test_program_from_dict_with_StateVector_fock_amplitude_map_preparation():
    amplitude_map = {(0,): 0.5, (1,): 0.5}

    instructions_dict = {
        "instructions": [
            {
                "type": "StateVector",
                "attributes": {
                    "constructor_kwargs": {"fock_amplitude_map": amplitude_map},
                    "modes": [0],
                },
            }
        ]
    }

    program = pq.Program.from_dict(instructions_dict)

    assert isinstance(program.instructions[0], pq.StateVector)
    assert program.instructions[0].params["fock_amplitude_map"] == amplitude_map
    assert program.instructions[0].modes == [0]


def test_program_from_dict_from_external_instruction():
    class FakeInstruction(pq.Instruction):
        def __init__(self, first_param, second_param):
            super().__init__(
                params=dict(
                    first_param=first_param,
                    second_param=second_param,
                ),
            )

    instructions_dict = {
        "instructions": [
            {
                "type": "FakeInstruction",
                "attributes": {
                    "constructor_kwargs": {
                        "first_param": "first_param_value",
                        "second_param": "second_param_value",
                    },
                    "modes": ["some", "modes"],
                },
            },
            {
                "type": "FakeInstruction",
                "attributes": {
                    "constructor_kwargs": {
                        "first_param": "2nd_instructions_1st_param_value",
                        "second_param": "2nd_instructions_2nd_param_value",
                    },
                    "modes": ["some", "other", "modes"],
                },
            },
        ]
    }

    program = pq.Program.from_dict(instructions_dict)

    assert isinstance(program.instructions[0], FakeInstruction)
    assert program.instructions[0].params == {
        "first_param": "first_param_value",
        "second_param": "second_param_value",
    }
    assert program.instructions[0].modes == ["some", "modes"]

    assert isinstance(program.instructions[1], FakeInstruction)
    assert program.instructions[1].params == {
        "first_param": "2nd_instructions_1st_param_value",
        "second_param": "2nd_instructions_2nd_param_value",
    }
    assert program.instructions[1].modes == ["some", "other", "modes"]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess

import numpy as np
import matplotlib.pyplot as plt

import piquasso as pq

import strawberryfields as sf


def gbs_nondisplaced_histogram_script():
    d = 5
    shots = 200

    # NOTE: In SF the measurement cutoff is 5, and couldn't be changed.
    pq_simulator = pq.GaussianSimulator(d=d, config=pq.Config(measurement_cutoff=5))

    with pq.Program() as pq_program:
        pq.Q(all) | pq.Squeezing(r=0.5)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

        pq.Q(0, 1, 2) | pq.ParticleNumberMeasurement()

    sf_program = sf.Program(d)
    sf_engine = sf.Engine(backend="gaussian")

    with sf_program.context as q:
        sf.ops.Sgate(0.5) | q[0]
        sf.ops.Sgate(0.5) | q[1]
        sf.ops.Sgate(0.5) | q[2]
        sf.ops.Sgate(0.5) | q[3]
        sf.ops.Sgate(0.5) | q[4]

        sf.ops.BSgate(0.0959408065906761, 0.06786053071484363) | (q[0], q[1])
        sf.ops.BSgate(0.7730047654405018, 1.453770233324797) | (q[2], q[3])
        sf.ops.BSgate(1.0152680371119776, 1.2863559998816205) | (q[1], q[2])
        sf.ops.BSgate(1.3205517879465705, 0.5236836466492961) | (q[3], q[4])
        sf.ops.BSgate(4.394480318177715, 4.481575657714487) | (q[0], q[1])
        sf.ops.BSgate(2.2300919706807534, 1.5073556513699888) | (q[2], q[3])
        sf.ops.BSgate(2.2679037068773673, 1.9550229282085838) | (q[1], q[2])
        sf.ops.BSgate(3.340269832485504, 3.289367083610399) | (q[3], q[4])

        sf.ops.MeasureFock() | (q[0], q[1], q[2])

    pq_results = np.array(pq_simulator.execute(pq_program, shots=shots).samples)
    sf_results = sf_engine.run(sf_program, shots=shots).samples

    n_bins = 20

    fig, axs = plt.subplots(1, 2, sharey=True, tight_layout=True)

    axs[0].hist(pq_results, bins=n_bins)
    axs[1].hist(sf_results, bins=n_bins)

    fig.savefig("histogram.png")

    subprocess.call(("xdg-open", "histogram.png"))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some of the code has been copyied from
`https://strawberryfields.ai/photonics/demos/run_gate_synthesis.html`.
"""

import pytest

import numpy as np

import tensorflow as tf

import piquasso as pq
from piquasso import cvqnn


np.set_printoptions(suppress=True, linewidth=200)


@pytest.fixture
def cutoff():
    return 20


@pytest.fixture
def layer_count():
    return 3


@pytest.fixture
def d():
    return 2


@pytest.fixture
def weights(layer_count, d):
    active_sd = 0.01
    passive_sd = 0.1

    M = int(d * (d - 1)) + max(1, d - 1)

    int1_weights = tf.random.normal(shape=[layer_count, M], stddev=passive_sd)
    s_weights = tf.random.normal(shape=[layer_count, d], stddev=active_sd)
    int2_weights = tf.random.normal(shape=[layer_count, M], stddev=passive_sd)
    dr_weights = tf.random.normal(shape=[layer_count, d], stddev=active_sd)
    dp_weights = tf.random.normal(shape=[layer_count, d], stddev=passive_sd)
    k_weights = tf.random.normal(shape=[layer_count, d], stddev=active_sd)

    weights = tf.cast(
        tf.concat(
            [int1_weights, s_weights, int2_weights, dr_weights, dp_weights, k_weights],
            axis=1,
        ),
        dtype=tf.float64,
    )

    weights = tf.Variable(weights)

    return weights


def piquasso_benchmark(benchmark, weights, cutoff):
    connector = pq.TensorflowConnector(decorate_with=tf.function)
    benchmark(lambda: _calculate_piquasso_results(weights, cutoff, connector))


def strawberryfields_benchmark(benchmark, weights, cutoff, sf):
    benchmark(lambda: _calculate_strawberryfields_results(weights, cutoff, sf))


def test_state_vector_and_jacobian(weights, cutoff, sf):
    pq_state_vector, pq_jacobian = _calculate_piquasso_results(
        weights, cutoff, pq.TensorflowConnector(decorate_with=tf.function)
    )
    sf_state_vector, sf_jacobian = _calculate_strawberryfields_results(
        weights, cutoff, sf
    )

    assert np.sum(np.abs(pq_state_vector - sf_state_vector) ** 2) < 1e-10
    assert np.sum(np.abs(pq_jacobian - sf_jacobian) ** 2) < 1e-10


def _pq_state_vector(weights, cutoff, connector):
    d = cvqnn.get_number_of_modes(weights.shape[1])

    simulator = pq.PureFockSimulator(
        d=d,
        config=pq.Config(cutoff=cutoff),
        connector=connector,
    )

    program = cvqnn.create_program(weights)

    state = simulator.execute(program).state

    return state.get_tensor_representation()


@tf.function
def _calculate_piquasso_results(weights, cutoff, connector):
    with tf.GradientTape() as tape:
        state_vector = _pq_state_vector(weights, cutoff, connector)

    return state_vector, tape.jacobian(state_vector, weights)


def _calculate_strawberryfields_results(weights, cutoff, sf):
    layer_count = weights.shape[0]
    d = cvqnn.get_number_of_modes(weights.shape[1])

    eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff})
    qnn = sf.Program(d)

    num_params = np.prod(weights.shape)

    sf_params = np.arange(num_params).reshape(weights.shape).astype(str)
    sf_params = np.array([qnn.params(*i) for i in sf_params])

    with qnn.context as q:
        for k in range(layer_count):
            _sf_layer(sf_params[k], q, sf)

    with tf.GradientTape() as tape:
        mapping = {
            p.name: w for p, w in zip(sf_params.flatten(), tf.reshape(weights, [-1]))
        }

        state = eng.run(qnn, args=mapping).state
        state_vector = state.ket()

    return state_vector, tape.jacobian(state_vector, weights)


def _sf_interferometer(params, q, sf):
    N = len(q)
    theta = params[: N * (N - 1) // 2]
    phi = params[N * (N - 1) // 2 : N * (N - 1)]
    rphi = params[-N + 1 :]

    if N == 1:
        sf.ops.Rgate(rphi[0]) | q[0]
        return

    n = 0

    for j in range(N):
        for k, (q1, q2) in enumerate(zip(q[:-1], q[1:])):
            if (j + k) % 2 != 1:
                sf.ops.BSgate(theta[n], phi[n]) | (q1, q2)
                n += 1

    for i in range(max(1, N - 1)):
        sf.ops.Rgate(rphi[i]) | q[i]


def _sf_layer(params, q, sf):
    N = len(q)
    M = int(N * (N - 1)) + max(1, N - 1)

    int1 = params[:M]
    s = params[M : M + N]
    int2 = params[M + N : 2 * M + N]
    dr = params[2 * M + N : 2 * M + 2 * N]
    dp = params[2 * M + 2 * N : 2 * M + 3 * N]
    k = params[2 * M + 3 * N : 2 * M + 4 * N]

    _sf_interferometer(int1, q, sf)

    for i in range(N):
        sf.ops.Sgate(s[i]) | q[i]

    _sf_interferometer(int2, q, sf)

    for i in range(N):
        sf.ops.Dgate(dr[i], dp[i]) | q[i]
        sf.ops.Kgate(k[i]) | q[i]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Package containing hafnian and loop hafnian calculations.

Most of the code in this package are translated versions of the PiquassoBoost C++ code
from https://github.com/Budapest-Quantum-Computing-Group/piquassoboost.

Moreover, the algorithms are enhanced by factoring in repetitions according to
https://arxiv.org/abs/2108.01622.
"""

from .plain_hafnian import hafnian_with_reduction, hafnian_with_reduction_batch

from .loop_hafnian import (
    loop_hafnian_with_reduction,
    loop_hafnian_with_reduction_batch,
)


__all__ = [
    "hafnian_with_reduction",
    "hafnian_with_reduction_batch",
    "loop_hafnian_with_reduction",
    "loop_hafnian_with_reduction_batch",
]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import numpy as np

from functools import partial

import piquasso as pq

from scipy.special import comb

from scipy.stats import unitary_group

from pathlib import Path

from scipy.linalg import polar, coshm, sinhm, logm

from piquasso._simulators.connectors import NumpyConnector


@pytest.fixture(autouse=True)
def _set_printoptions_for_debugging():
    np.set_printoptions(suppress=True, linewidth=200, precision=6)


@pytest.fixture
def generate_symmetric_matrix():
    def func(N):
        A = np.random.rand(N, N)

        return A + A.T

    return func


@pytest.fixture
def generate_complex_symmetric_matrix(generate_symmetric_matrix):
    def func(N):
        real = generate_symmetric_matrix(N)
        imaginary = generate_symmetric_matrix(N)

        return real + 1j * imaginary

    return func


@pytest.fixture
def generate_unitary_matrix():
    def func(N):
        if N == 1:
            return np.array([[np.exp(np.random.rand() * 2 * np.pi * 1j)]])

        return np.array(unitary_group.rvs(N), dtype=complex)

    return func


@pytest.fixture
def generate_random_positive_definite_matrix():
    def func(N):
        A = np.random.rand(N, N)
        return A @ A.transpose()

    return func


@pytest.fixture
def generate_hermitian_matrix(generate_unitary_matrix):
    def func(N):
        U = generate_unitary_matrix(N)

        return 1j * logm(U)

    return func


@pytest.fixture
def generate_skew_symmetric_matrix(generate_unitary_matrix):
    def func(N):
        A = np.random.rand(N, N) + 1j * np.random.rand(N, N)

        return A - A.T

    return func


@pytest.fixture
def generate_gaussian_transform(
    generate_complex_symmetric_matrix, generate_unitary_matrix
):
    def func(d):
        squeezing_matrix = generate_complex_symmetric_matrix(d)
        U, r = polar(squeezing_matrix)

        global_phase = generate_unitary_matrix(d)
        passive = global_phase @ coshm(r)
        active = global_phase @ sinhm(r) @ U.conj()

        return pq.GaussianTransform(passive=passive, active=active)

    return func


@pytest.fixture
def AssetHandler(request):
    class _AssetHandler:
        def __init__(self):
            testfile = request.module.__file__
            self._asset_dir = Path(testfile).parent / "assets"

            if not os.path.exists(self._asset_dir):
                os.makedirs(self._asset_dir)

            name = request.function.__name__ + "."

            if request.cls:
                name = request.cls.__name__ + "." + name

            self._asset_id = Path(os.path.splitext(testfile)[0]).name + "." + name

        def load(self, asset: str, loader: any = None):
            loader = loader or self._load_numpy_array

            filename = self._resolve_file_for_load(asset)

            return loader(filename)

        def save(self, asset: str, obj: any):
            filename = self._resolve_file_for_save(asset)

            self._write_repr(filename, obj)

        def _resolve_file_for_load(self, asset):
            pattern = self._asset_id + asset

            files = list(Path(self._asset_dir).glob(pattern))

            assert len(files) == 1, (
                f"There should be at least one asset file named '{pattern}' "
                f"under '{self._asset_dir}'."
            )

            return files[0]

        @staticmethod
        def _load_numpy_array(filename: str):
            with open(filename, "r") as f:
                data = f.read()

            # NOTE: This is needed for the evaluation with `eval` below, where `array`'s
            # are called.
            from numpy import array  # noqa: F401

            return eval(data)

        def _resolve_file_for_save(self, asset):
            return self._asset_dir / (self._asset_id + asset)

        @staticmethod
        def _write_repr(filename, obj):
            with open(filename, "w") as outfile:
                outfile.write(repr(obj))

    return _AssetHandler


@pytest.fixture
def assets(AssetHandler):
    return AssetHandler()


@pytest.fixture
def gaussian_state_assets(AssetHandler):
    class GaussianStateAssetHandler(AssetHandler):
        def __init__(self):
            super().__init__()

            self._asset = "expected_state"

        def load(self):
            filename = self._resolve_file_for_load(asset=self._asset)

            return self._load_gaussian_state(filename)

        def save(self, obj: any):
            state_map = {
                "class": "GaussianState",
                "properties": {
                    "xpxp_mean_vector": obj.xpxp_mean_vector,
                    "xpxp_covariance_matrix": obj.xpxp_covariance_matrix,
                },
            }
            np.set_printoptions(precision=16)

            filename = self._resolve_file_for_save(self._asset)

            np.set_printoptions(precision=8)

            self._write_repr(filename, state_map)

        @staticmethod
        def _load_gaussian_state(filename):
            """
            Note:
                Do not use the `eval` approach ever in production code, it's insecure!
            """
            with open(filename, "r") as f:
                data = f.read()

            # NOTE: This is needed for the evaluation with `eval` below, where `array`'s
            # are called.
            from numpy import array  # noqa: F401

            properties = eval(data)["properties"]

            d = len(properties["xpxp_mean_vector"]) // 2

            state = pq.GaussianState(d=d, connector=NumpyConnector())

            state.xpxp_mean_vector = properties["xpxp_mean_vector"]
            state.xpxp_covariance_matrix = properties["xpxp_covariance_matrix"]

            return state

    return GaussianStateAssetHandler()


def _get_click_distribution(d, n):
    dist = np.empty(shape=(d,))
    for c in range(1, d + 1):
        dist[c - 1] = comb(d, c) * comb(n - 1, n - c) / comb(n + d - 1, n)

    return dist


@pytest.fixture(scope="session")
def generate_random_fock_state():
    def func(d, n):
        occupation_numbers = np.zeros(shape=(d,), dtype=int)

        modes = tuple(range(d))
        c = d
        k = n

        while k > 0:
            dist = _get_click_distribution(c, k)

            c = np.random.choice(tuple(range(1, c + 1)), p=dist)
            k -= c

            modes = np.random.choice(modes, size=c, replace=False)

            occupation_numbers[modes] += 1

        return occupation_numbers

    return func


def pytest_addoption(parser):
    parser.addoption(
        "--skip-tensorflow",
        action="store_true",
        help="Skips tests using TensorFlow.",
    )


@pytest.fixture
def tf(request):
    skip_tf = request.config.getoption("--skip-tensorflow")

    if skip_tf:
        pytest.skip(
            reason=(
                "This test is skipped, as it uses TensorFlow, and the "
                "'--skip-tensorflow' option is enabled."
            )
        )

    import tensorflow

    return tensorflow


@pytest.fixture
def tensorflow_connector(tf):
    return pq.TensorflowConnector()


@pytest.fixture
def tensorflow_connector_tf_function(tf):
    return pq.TensorflowConnector(decorate_with=tf.function)


@pytest.fixture
def PureFockSimulator_with_tensorflow(tensorflow_connector):
    return partial(pq.PureFockSimulator, connector=tensorflow_connector)


@pytest.fixture
def PureFockSimulator_with_tensorflow_tf_function(tensorflow_connector_tf_function):
    return partial(pq.PureFockSimulator, connector=tensorflow_connector_tf_function)


@pytest.fixture
def PureFockSimulator_with_jax():
    return partial(pq.PureFockSimulator, connector=pq.JaxConnector())



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Some of the code has been copyied from
`https://strawberryfields.ai/photonics/demos/run_gate_synthesis.html`.
"""

import numpy as np

import tensorflow as tf
import strawberryfields as sf

import piquasso as pq
from piquasso import cvqnn

import time

import json


filename = f"cvqnn_{int(time.time())}.json"

# Set TensorFlow to use only the CPU
tf.config.set_visible_devices([], "GPU")

# Verify devices
# This step is optional, just to confirm that GPU is not being used.
physical_devices = tf.config.list_physical_devices()
print("Available devices:", physical_devices)


ACTIVE_VAR = 0.1
PASSIVE_VAR = 1.0


tf.get_logger().setLevel("ERROR")
np.set_printoptions(suppress=True, linewidth=200)


def measure_graph_size(f, *args):
    if not hasattr(f, "get_concrete_function"):
        return 0

    g = f.get_concrete_function(*args).graph

    return len(g.as_graph_def().node)


def _pq_loss(weights, cutoff, connector):
    d = cvqnn.get_number_of_modes(weights.shape[1])

    simulator = pq.PureFockSimulator(
        d=d,
        config=pq.Config(cutoff=cutoff),
        connector=connector,
    )

    program = cvqnn.create_program(weights)

    state_vector = simulator.execute(program).state.state_vector

    return tf.math.reduce_mean(
        (
            tf.ones_like(state_vector) / np.sqrt(len(state_vector))
            - tf.math.real(state_vector)
        )
        ** 2
    )


def _calculate_piquasso_results(weights, cutoff, connector):
    with tf.GradientTape() as tape:
        loss = _pq_loss(weights, cutoff, connector)

    return loss, tape.gradient(loss, weights)


def _calculate_strawberryfields_results(weights, cutoff):
    layer_count = weights.shape[0]
    d = cvqnn.get_number_of_modes(weights.shape[1])

    eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff})
    qnn = sf.Program(d)

    num_params = np.prod(weights.shape)

    sf_params = np.arange(num_params).reshape(weights.shape).astype(str)
    sf_params = np.array([qnn.params(*i) for i in sf_params])

    with qnn.context as q:
        for k in range(layer_count):
            _sf_layer(sf_params[k], q)

    with tf.GradientTape() as tape:
        mapping = {
            p.name: w for p, w in zip(sf_params.flatten(), tf.reshape(weights, [-1]))
        }

        state = eng.run(qnn, args=mapping).state
        state_vector = state.ket()
        loss = tf.math.reduce_mean(
            (
                tf.ones_like(state_vector) / np.sqrt(len(state_vector))
                - tf.math.real(state_vector)
            )
            ** 2
        )

    return loss, tape.gradient(loss, weights)


def _sf_interferometer(params, q):
    N = len(q)
    theta = params[: N * (N - 1) // 2]
    phi = params[N * (N - 1) // 2 : N * (N - 1)]
    rphi = params[-N + 1 :]

    if N == 1:
        sf.ops.Rgate(rphi[0]) | q[0]
        return

    n = 0

    for j in range(N):
        for k, (q1, q2) in enumerate(zip(q[:-1], q[1:])):
            if (j + k) % 2 != 1:
                sf.ops.BSgate(theta[n], phi[n]) | (q1, q2)
                n += 1

    for i in range(max(1, N - 1)):
        sf.ops.Rgate(rphi[i]) | q[i]


def _sf_layer(params, q):
    N = len(q)
    M = int(N * (N - 1)) + max(1, N - 1)

    int1 = params[:M]
    s = params[M : M + N]
    int2 = params[M + N : 2 * M + N]
    dr = params[2 * M + N : 2 * M + 2 * N]
    dp = params[2 * M + 2 * N : 2 * M + 3 * N]
    k = params[2 * M + 3 * N : 2 * M + 4 * N]

    _sf_interferometer(int1, q)

    for i in range(N):
        sf.ops.Sgate(s[i]) | q[i]

    _sf_interferometer(int2, q)

    for i in range(N):
        sf.ops.Dgate(dr[i], dp[i]) | q[i]
        sf.ops.Kgate(k[i]) | q[i]


if __name__ == "__main__":
    cutoff = 6
    layer_count = 4
    NUMBER_OF_ITERATIONS = 1

    run_pq = True
    run_sf = False

    results = {"pq_times": [], "sf_times": []}

    for d in range(2, 25 + 1):
        print(d)
        weights = tf.Variable(
            pq.cvqnn.generate_random_cvqnn_weights(
                layer_count=layer_count,
                d=d,
                active_var=ACTIVE_VAR,
                passive_var=PASSIVE_VAR,
            )
        )

        connector = pq.TensorflowConnector()

        # Warmup
        if run_pq:
            _calculate_piquasso_results(
                weights,
                cutoff,
                connector,
            )
            for i in range(NUMBER_OF_ITERATIONS):
                weights = tf.Variable(
                    pq.cvqnn.generate_random_cvqnn_weights(
                        layer_count=layer_count,
                        d=d,
                        active_var=ACTIVE_VAR,
                        passive_var=PASSIVE_VAR,
                    )
                )
                start_time = time.time()
                _calculate_piquasso_results(weights, cutoff, connector)
                runtime = time.time() - start_time
                print("PQ RUNTIME:", runtime)
                results["pq_times"].append(runtime)

                with open(filename, "w") as f:
                    json.dump(results, f)

        if run_sf:
            _calculate_strawberryfields_results(weights, cutoff)

            for i in range(NUMBER_OF_ITERATIONS):
                weights = tf.Variable(
                    pq.cvqnn.generate_random_cvqnn_weights(
                        layer_count=layer_count,
                        d=d,
                        active_var=ACTIVE_VAR,
                        passive_var=PASSIVE_VAR,
                    )
                )
                start_time = time.time()
                _calculate_strawberryfields_results(weights, cutoff)
                runtime = time.time() - start_time
                print("SF RUNTIME:", runtime)
                results["sf_times"].append(runtime)

                with open(filename, "w") as f:
                    json.dump(results, f)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numba as nb
import numpy as np

from numba import complex128, int64

from piquasso._math.combinatorics import comb

from .powtrace import calculate_power_traces_loop

from .utils import match_occupation_numbers, ix_, get_kept_edges
from .loop_corrections import calculate_loop_corrections


@nb.njit(cache=True)
def _get_scale_factor(matrix):
    r"""
    Scales the matrix for better accuracy.
    """

    dim = matrix.shape[0]

    if dim <= 10:
        return 1.0

    scale_factor = np.sum(np.abs(matrix))

    scale_factor = scale_factor / dim**2 / np.sqrt(2.0)

    return scale_factor


@nb.njit(cache=True)
def _scale_matrix_and_diagonal(matrix, diagonal, scale_factor):
    inverse_scale_factor = 1 / scale_factor

    dim = len(diagonal)

    for row_idx in range(dim):
        for col_idx in range(dim):
            if col_idx == row_idx:
                sqrt_inverse_scale_factor = np.sqrt(inverse_scale_factor)
                diagonal[row_idx] = diagonal[row_idx] * sqrt_inverse_scale_factor

            matrix[row_idx, col_idx] = matrix[row_idx, col_idx] * inverse_scale_factor

    return matrix, diagonal


@nb.njit(cache=True)
def _calc_B_and_reduced_diagonals(matrix, diagonal, delta):
    r"""
    Creates the input data required to calculate power traces and loop corrections
    according to `delta`.

    Args:
        matrix (np.ndarray): The input matrix.
        diagonal (np.ndarray): The displacement term.
        delta (np.ndarray): Elements required in $X_{\delta}$ from Eq. (B11) from
            https://arxiv.org/abs/2108.01622.

    Returns:
        Tuple: The input data required to calculate power traces and loop corrections.
    """
    nonzero_indices = np.nonzero(delta)[0]
    nonzero_dim_over_2 = len(nonzero_indices)

    nonzero_dim = 2 * nonzero_dim_over_2

    B = np.empty((nonzero_dim, nonzero_dim), dtype=matrix.dtype)
    reduced_diagonal_left = np.empty(nonzero_dim, dtype=matrix.dtype)
    reduced_diagonal_right = np.empty(nonzero_dim, dtype=matrix.dtype)

    for i in range(nonzero_dim_over_2):
        even = 2 * i
        odd = 2 * i + 1

        delta_i = delta[nonzero_indices[i]]

        source_col_even = 2 * nonzero_indices[i] + 1
        source_col_odd = 2 * nonzero_indices[i]

        for row_idx in range(nonzero_dim):
            source_row = 2 * nonzero_indices[row_idx // 2] + row_idx % 2

            B[row_idx, even] = delta_i * matrix[source_row, source_col_even]
            B[row_idx, odd] = delta_i * matrix[source_row, source_col_odd]

        reduced_diagonal_right[even] = diagonal[source_col_odd]
        reduced_diagonal_right[odd] = diagonal[source_col_even]

        reduced_diagonal_left[even] = delta_i * diagonal[source_col_even]
        reduced_diagonal_left[odd] = delta_i * diagonal[source_col_odd]

    return B, reduced_diagonal_left, reduced_diagonal_right


@nb.njit(cache=True)
def _calc_f_loop(traces, loop_corrections):
    r"""Calcualates `f` from Appendix B.1 from https://arxiv.org/abs/2108.01622.

    This function uses the previously calculated power traces and loop corrections.

    Args:
        traces (numpy.ndarray): The power traces.
        loop_corrections (numpy.ndarray): Loop corrections.
    """

    dim_over_2 = len(traces)
    dim = 2 * dim_over_2

    aux0 = np.zeros(dim_over_2 + 1, dtype=traces.dtype)
    aux1 = np.zeros(dim_over_2 + 1, dtype=traces.dtype)

    aux0[0] = 1.0

    data = [aux0, aux1]

    p_aux0 = 0
    p_aux1 = 1

    for idx in range(1, dim_over_2 + 1):
        factor = traces[idx - 1] / (2.0 * idx) + loop_corrections[idx - 1] * 0.5

        powfactor = 1.0

        if idx % 2 == 1:
            p_aux0 = 0
            p_aux1 = 1
        else:
            p_aux0 = 1
            p_aux1 = 0

        data[p_aux1] = np.copy(data[p_aux0])

        for jdx in range(1, dim // (2 * idx) + 1):
            powfactor = powfactor * factor / jdx

            for kdx in range(idx * jdx, dim_over_2 + 1):
                data[p_aux1][kdx] += data[p_aux0][kdx - idx * jdx] * powfactor

    return data[p_aux1]


@nb.njit(cache=True)
def _extend(matrix_orig, diagonal_orig, occupation_numbers):
    dim = len(matrix_orig)
    new_matrix = np.empty((dim + 1, dim + 1), dtype=matrix_orig.dtype)
    new_matrix[1:, 1:] = matrix_orig
    new_matrix[0, 0] = 1.0
    new_matrix[1:, 0] = 0.0
    new_matrix[0, 1:] = 0.0

    d = len(occupation_numbers)

    new_occupation_numbers = np.empty(d + 1, dtype=occupation_numbers.dtype)
    new_diagonal = np.empty(d + 1, dtype=diagonal_orig.dtype)

    new_occupation_numbers[0] = 1
    new_diagonal[0] = 1.0

    for i in range(d):
        new_occupation_numbers[i + 1] = occupation_numbers[i]
        new_diagonal[i + 1] = diagonal_orig[i]

    return new_matrix, new_diagonal, new_occupation_numbers


@nb.njit(cache=True, parallel=True)
def loop_hafnian_with_reduction(matrix_orig, diagonal_orig, occupation_numbers):
    r"""
    Calculates the loop hafnian of the input matrix using the power trace algorithm
    with Glynn-type iterations.

    The algorithm is enhanced by factoring in repetitions according to
    https://arxiv.org/abs/2108.01622.
    """

    n = sum(occupation_numbers)

    diagonal_orig = diagonal_orig.astype(matrix_orig.dtype)

    if n == 0:
        return 1.0
    elif n % 2 == 1:
        # Handling the odd case by extending the matrix with a 1 to be even.
        #
        # TODO: This is not the best handling of the odd case, and we can definitely
        # squeeze out a bit more performance if needed.
        matrix_orig, diagonal_orig, occupation_numbers = _extend(
            matrix_orig, diagonal_orig, occupation_numbers
        )

    all_edges, edge_indices = match_occupation_numbers(occupation_numbers)

    matrix = ix_(matrix_orig, edge_indices, edge_indices)
    diagonal = diagonal_orig[edge_indices]

    scale_factor = _get_scale_factor(matrix)
    matrix, diagonal = _scale_matrix_and_diagonal(matrix, diagonal, scale_factor)

    dim_over_2 = sum(all_edges)
    number_of_reps = len(all_edges)

    result = 0.0

    comb_cache = {}

    for n in range(max(all_edges) + 1):
        for k in range(n + 1):
            comb_cache[(n, k)] = comb(n, k)

    size = np.prod(all_edges + 1) // 2

    for permutation_idx in nb.prange(size):
        kept_edges = get_kept_edges(all_edges, permutation_idx)
        fact = False
        combinatorial_factor = 1.0

        delta = np.empty_like(all_edges)

        for i in range(number_of_reps):
            no_of_edges = all_edges[i]
            no_of_kept_edges = kept_edges[i]

            fact ^= (no_of_edges - no_of_kept_edges) % 2
            comb_input = (no_of_edges, no_of_kept_edges)
            combinatorial_factor *= comb_cache[comb_input]

            delta[i] = 2 * no_of_kept_edges - no_of_edges

        prefactor = (-1 if fact else 1) * combinatorial_factor

        fs = _calculate_f_loop_from_matrix(matrix, diagonal, delta, dim_over_2)

        summand = prefactor * fs[dim_over_2]

        result += summand

    result *= scale_factor**dim_over_2

    result /= 1 << (dim_over_2 - 1)

    return result


@nb.njit(cache=True)
def _prepare_data(matrix_orig, diagonal_orig, occupation_numbers_orig, cutoff):
    int_dtype = occupation_numbers_orig.dtype

    particle_number_sum = sum(occupation_numbers_orig)

    result_size = (cutoff - 1) // 2
    result_size_odd = (cutoff - 2) // 2

    occupation_numbers_orig_copy = np.copy(occupation_numbers_orig)

    if particle_number_sum % 2 == 0:
        all_edges_orig, edge_indices_orig = match_occupation_numbers(
            occupation_numbers_orig
        )
        matrix = matrix_orig
        diagonal = diagonal_orig

        all_edges = np.concatenate(
            (all_edges_orig, np.array((result_size,), dtype=int_dtype))
        )
        last_col_idx = len(matrix) - 1
        edge_indices = np.concatenate(
            (
                edge_indices_orig,
                np.array((last_col_idx, last_col_idx), dtype=int_dtype),
            ),
        )

        occupation_numbers_orig_copy[-1] += 1

        matrix_odd, diagonal_odd, _ = _extend(
            matrix_orig, diagonal_orig, occupation_numbers_orig_copy
        )

        all_edges_odd = np.concatenate(
            (
                np.array((1,), dtype=int_dtype),
                all_edges_orig,
                np.array((result_size_odd,), dtype=int_dtype),
            )
        )
        last_col_idx = len(matrix_odd) - 1
        edge_indices_odd = np.concatenate(
            (
                np.array((0, last_col_idx), dtype=int_dtype),
                edge_indices_orig + 1,
                np.array((last_col_idx, last_col_idx), dtype=int_dtype),
            )
        )

    else:
        matrix, diagonal, occupation_numbers = _extend(
            matrix_orig, diagonal_orig, occupation_numbers_orig
        )
        all_edges_orig, edge_indices_orig = match_occupation_numbers(occupation_numbers)
        all_edges = np.concatenate(
            (all_edges_orig, np.array((result_size,), dtype=int_dtype))
        )
        last_col_idx = len(matrix) - 1
        edge_indices = np.concatenate(
            (
                edge_indices_orig,
                np.array((last_col_idx, last_col_idx), dtype=int_dtype),
            ),
        )

        occupation_numbers_orig_copy[-1] += 1

        matrix_odd = matrix_orig
        diagonal_odd = diagonal_orig

        all_edges_odd, edge_indices_odd = match_occupation_numbers(
            occupation_numbers_orig_copy
        )
        all_edges_odd = np.concatenate(
            (all_edges_odd, np.array((result_size_odd,), dtype=int_dtype))
        )
        last_col_idx = len(matrix_odd) - 1
        edge_indices_odd = np.concatenate(
            (edge_indices_odd, np.array((last_col_idx, last_col_idx), dtype=int_dtype))
        )

    matrix = ix_(matrix, edge_indices, edge_indices)
    diagonal = diagonal[edge_indices]

    matrix_odd = ix_(matrix_odd, edge_indices_odd, edge_indices_odd)
    diagonal_odd = diagonal_odd[edge_indices_odd]

    return matrix, diagonal, matrix_odd, diagonal_odd, all_edges, all_edges_odd


@nb.njit(cache=True)
def _concatenate_results(
    result, result_odd, scale_factor, scale_factor_odd, dim_over_2, dim_over_2_odd
):
    dim_even = len(result)
    dim_odd = len(result_odd)
    concat_result = np.empty(dim_even + dim_odd, dtype=result.dtype)

    for i in range(dim_even):
        concat_result[2 * i] = (
            result[i]
            / (1 << (dim_over_2 - dim_even + 1 + i))
            * scale_factor ** (dim_over_2 - dim_even + 1 + i)
        )

    for i in range(dim_odd):
        concat_result[2 * i + 1] = (
            result_odd[i]
            / (1 << (dim_over_2_odd - dim_odd + 1 + i))
            * scale_factor_odd ** (dim_over_2_odd - dim_odd + 1 + i)
        )

    return concat_result


@nb.njit(cache=True)
def _calculate_aux_data(all_edges, kept_edges, comb_cache):
    comb_factor = 1.0
    fact = False

    number_of_reps = len(all_edges)

    delta = np.empty_like(all_edges)

    for i in range(number_of_reps):
        no_of_edges = all_edges[i]
        no_of_kept_edges = kept_edges[i]

        fact ^= (no_of_edges - no_of_kept_edges) % 2

        if i != number_of_reps - 1:
            comb_input = (no_of_edges, no_of_kept_edges)
            comb_factor *= comb_cache[comb_input]

        delta[i] = 2 * no_of_kept_edges - no_of_edges

    return delta, fact, comb_factor


@nb.njit(cache=True)
def _calculate_f_loop_from_matrix(matrix, diagonal, delta, dim_over_2):
    B, reduced_diagonal_left, reduced_diagonal_right = _calc_B_and_reduced_diagonals(
        matrix, diagonal, delta
    )

    traces = calculate_power_traces_loop(
        reduced_diagonal_right, reduced_diagonal_left, B, dim_over_2
    )

    loop_corrections = calculate_loop_corrections(
        reduced_diagonal_right, reduced_diagonal_left, B, dim_over_2
    )

    return _calc_f_loop(traces, loop_corrections)


@nb.njit(cache=True)
def _calculate_summand(fs, result_size, kept_edges, fact, comb_factor, comb_cache):
    dim_over_2 = len(fs) - 1

    summand = np.zeros(result_size + 1, dtype=fs.dtype)
    for idx in range(kept_edges[-1], result_size + 1):
        prefactor = (
            (-1 if fact ^ ((idx - result_size) % 2) else 1)
            * comb_factor
            * comb_cache[idx, kept_edges[-1]]
        )

        if idx >= result_size - kept_edges[-1]:
            prefactor *= (
                1
                + (-1 if (result_size - idx) % 2 else 1)
                * comb_cache[idx, result_size - kept_edges[-1]]
                / comb_cache[idx, kept_edges[-1]]
            )

        summand[idx] = prefactor * fs[dim_over_2 - result_size + idx]

    return summand


@nb.njit(
    complex128[:](complex128[:, :], complex128[:], int64[:], int64),
    cache=True,
    parallel=True,
)
def loop_hafnian_with_reduction_batch(
    matrix_orig, diagonal_orig, occupation_numbers_orig, cutoff
):
    r"""
    Calculates the loop hafnian of the input matrix using the power trace algorithm with
    Glynn-type iterations with batching, as described in
    https://arxiv.org/abs/2108.01622.
    """
    particle_number_sum = sum(occupation_numbers_orig)

    matrix, diagonal, matrix_odd, diagonal_odd, all_edges, all_edges_odd = (
        _prepare_data(
            matrix_orig,
            diagonal_orig,
            occupation_numbers_orig,
            cutoff,
        )
    )

    scale_factor = _get_scale_factor(matrix)
    scale_factor_odd = _get_scale_factor(matrix_odd)

    matrix, diagonal = _scale_matrix_and_diagonal(matrix, diagonal, scale_factor)
    matrix_odd, diagonal_odd = _scale_matrix_and_diagonal(
        matrix_odd, diagonal_odd, scale_factor_odd
    )

    comb_cache = nb.typed.Dict()
    for n in range(max(max(all_edges_odd), max(all_edges)) + 1):
        for k in range(n + 1):
            comb_cache[(n, k)] = comb(n, k)

    dim_over_2 = sum(all_edges)
    dim_over_2_odd = sum(all_edges_odd)

    size = np.prod(all_edges + 1) // 2
    size_odd = np.prod(all_edges_odd + 1) // 2

    result_size = (cutoff - 1) // 2
    result_size_odd = (cutoff - 2) // 2

    result = np.zeros(result_size + 1, dtype=matrix.dtype)
    result_odd = np.zeros(result_size_odd + 1, dtype=matrix.dtype)

    for permutation_idx in nb.prange(size + size_odd):
        if permutation_idx < size:
            kept_edges = get_kept_edges(all_edges, permutation_idx)

            delta, fact, comb_factor = _calculate_aux_data(
                all_edges, kept_edges, comb_cache
            )

            fs = _calculate_f_loop_from_matrix(matrix, diagonal, delta, dim_over_2)

            summand = _calculate_summand(
                fs, result_size, kept_edges, fact, comb_factor, comb_cache
            )

            result += summand
        else:
            kept_edges = get_kept_edges(all_edges_odd, permutation_idx - size)

            delta, fact, comb_factor = _calculate_aux_data(
                all_edges_odd, kept_edges, comb_cache
            )

            fs = _calculate_f_loop_from_matrix(
                matrix_odd, diagonal_odd, delta, dim_over_2_odd
            )

            summand = _calculate_summand(
                fs, result_size_odd, kept_edges, fact, comb_factor, comb_cache
            )

            result_odd += summand

    concat_result = _concatenate_results(
        result, result_odd, scale_factor, scale_factor_odd, dim_over_2, dim_over_2_odd
    )

    if particle_number_sum == 0:
        concat_result[0] = 1.0

    return concat_result



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from scipy.stats import chisquare, unitary_group


@pytest.mark.monkey
def test_gaussian_boson_sampling_chi_square_hypothesis_test():
    d = 3
    shots = 10000

    with pq.Program() as gaussian_boson_sampling:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(np.random.rand() * 0.1) | pq.Displacement(
                np.random.normal(loc=0.0, scale=0.5)
            )

        pq.Q(all) | pq.Interferometer(unitary_group.rvs(d))

        pq.Q(all) | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(
        d=d,
        connector=pq.NumpyConnector(),
        config=pq.Config(
            measurement_cutoff=10,
        ),
    )
    result = simulator.execute(gaussian_boson_sampling, shots=shots)
    samples = result.samples
    state = result.state

    samples_as_list = [tuple(x) for x in samples]

    sample_set = set(samples_as_list)

    f_obs = []
    f_exp = []

    for sample in sample_set:
        f_obs.append(samples_as_list.count(sample) / shots)
        f_exp.append(state.get_particle_detection_probability(sample).real)

    f_obs = np.array(f_obs) / np.sum(f_obs)
    f_exp = np.array(f_exp) / np.sum(f_exp)

    test = chisquare(f_obs, f_exp)

    assert test.pvalue >= 0.05, (
        "Chi-square hypothesis test failed: the GBS samples differ from the expected "
        "distribution."
    )


@pytest.mark.monkey
def test_threshold_gaussian_boson_sampling_chi_square_hypothesis_test():
    d = 3
    shots = 10000

    with pq.Program() as gaussian_boson_sampling:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(np.random.rand() * 0.1) | pq.Displacement(
                np.random.normal(loc=0.0, scale=0.5)
            )

        pq.Q(all) | pq.Interferometer(unitary_group.rvs(d))

        pq.Q(all) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(
        d=d,
        connector=pq.NumpyConnector(),
        config=pq.Config(
            measurement_cutoff=10,
        ),
    )
    result = simulator.execute(gaussian_boson_sampling, shots=shots)
    samples = result.samples
    state = result.state

    samples_as_list = [tuple(x) for x in samples]

    sample_set = set(samples_as_list)

    f_obs = []
    f_exp = []

    for sample in sample_set:
        f_obs.append(samples_as_list.count(sample) / shots)
        f_exp.append(state.get_particle_detection_probability(sample).real)

    f_obs = np.array(f_obs) / np.sum(f_obs)
    f_exp = np.array(f_exp) / np.sum(f_exp)

    test = chisquare(f_obs, f_exp)

    assert test.pvalue >= 0.05, (
        "Chi-square hypothesis test failed: the GBS samples differ from the expected "
        "distribution."
    )


@pytest.mark.monkey
def test_threshold_gaussian_boson_sampling_torontonian_chi_square_hypothesis_test():
    d = 3
    shots = 10000

    with pq.Program() as gaussian_boson_sampling:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(np.random.rand() * 0.1) | pq.Displacement(
                np.random.normal(loc=0.0, scale=0.5)
            )

        pq.Q(all) | pq.Interferometer(unitary_group.rvs(d))

        pq.Q(all) | pq.ThresholdMeasurement()

    simulator = pq.GaussianSimulator(
        d=d,
        connector=pq.NumpyConnector(),
        config=pq.Config(
            measurement_cutoff=10,
            use_torontonian=True,
        ),
    )
    result = simulator.execute(gaussian_boson_sampling, shots=shots)
    samples = result.samples
    state = result.state

    samples_as_list = [tuple(x) for x in samples]

    sample_set = set(samples_as_list)

    f_obs = []
    f_exp = []

    for sample in sample_set:
        f_obs.append(samples_as_list.count(sample) / shots)
        f_exp.append(state.get_particle_detection_probability(sample).real)

    f_obs = np.array(f_obs) / np.sum(f_obs)
    f_exp = np.array(f_exp) / np.sum(f_exp)

    test = chisquare(f_obs, f_exp)

    assert test.pvalue >= 0.05, (
        "Chi-square hypothesis test failed: the GBS samples differ from the expected "
        "distribution."
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numba as nb
import numpy as np


@nb.njit(cache=True)
def labudde(A):
    """Call to determine the coefficients of the characteristic polynomial using the
    Algorithm 2 of LaBudde method.
    See [arXiv:1104.3769](https://arxiv.org/abs/1104.3769v1) for further details.
    """
    dim = A.shape[0]

    mtx = A.flatten()

    coeffs = np.zeros(dim * dim, dtype=A.dtype)

    coeffs[0] = -mtx[0]

    coeffs[dim] = coeffs[0] - mtx[dim + 1]

    coeffs[dim + 1] = mtx[0] * mtx[dim + 1] - mtx[1] * mtx[dim]

    beta_prods = np.zeros(dim, dtype=A.dtype)

    for idx in range(2, dim):
        beta_prods[0] = mtx[idx * dim + idx - 1]

        for prod_idx in range(1, idx):
            beta_prods[prod_idx] = (
                beta_prods[prod_idx - 1]
                * mtx[(idx - prod_idx) * dim + (idx - prod_idx - 1)]
            )

        coeffs[idx * dim] = coeffs[(idx - 1) * dim] - mtx[idx * dim + idx]

        for jdx in range(1, idx):
            sum = 0.0

            for mdx in range(1, jdx):
                sum += (
                    mtx[(idx - mdx) * dim + idx]
                    * beta_prods[mdx - 1]
                    * coeffs[(idx - mdx - 1) * dim + jdx - mdx - 1]
                )

            sum += mtx[(idx - jdx) * dim + idx] * beta_prods[jdx - 1]

            coeffs[idx * dim + jdx] = (
                coeffs[(idx - 1) * dim + jdx]
                - mtx[idx * dim + idx] * coeffs[(idx - 1) * dim + jdx - 1]
                - sum
            )

        sum = 0.0

        for mdx in range(1, idx):
            sum += (
                mtx[(idx - mdx) * dim + idx]
                * beta_prods[mdx - 1]
                * coeffs[(idx - mdx - 1) * dim + idx - mdx - 1]
            )

        coeffs[idx * dim + idx] = (
            -mtx[idx * dim + idx] * coeffs[(idx - 1) * dim + idx - 1]
            - sum
            - mtx[idx] * beta_prods[idx - 1]
        )

    for idx in range(dim, dim):
        coeffs[idx * dim] = coeffs[(idx - 1) * dim] - mtx[idx * dim + idx]

        if dim >= 2:
            beta_prods[0] = mtx[idx * dim + idx - 1]

            for prod_idx in range(1, idx):
                beta_prods[prod_idx] = (
                    beta_prods[prod_idx - 1]
                    * mtx[(idx - prod_idx) * dim + (idx - prod_idx - 1)]
                )

            for jdx in range(1, dim):
                sum = 0.0

                for mdx in range(1, jdx):
                    sum += (
                        mtx[(idx - mdx) * dim + idx]
                        * beta_prods[mdx - 1]
                        * coeffs[(idx - mdx - 1) * dim + jdx - mdx - 1]
                    )

                sum += mtx[(idx - jdx) * dim + idx] * beta_prods[jdx - 1]

                coeffs[idx * dim + jdx] = (
                    coeffs[(idx - 1) * dim + jdx]
                    - mtx[idx * dim + idx] * coeffs[(idx - 1) * dim + jdx - 1]
                    - sum
                )

    return coeffs.reshape((dim, dim))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np

from ...calculations import get_projection_operator_indices

from ..state import PureFockState

from piquasso.api.connector import BaseConnector


def project_to_subspace(
    state: PureFockState,
    *,
    subspace_basis: Tuple[int, ...],
    modes: Tuple[int, ...],
    normalization: float,
    connector: BaseConnector,
) -> None:
    projected_state_vector = _get_projected_state_vector(
        state=state, subspace_basis=subspace_basis, modes=modes, connector=connector
    )

    state.state_vector = projected_state_vector * normalization


def _get_projected_state_vector(
    state: PureFockState,
    *,
    subspace_basis: Tuple[int, ...],
    modes: Tuple[int, ...],
    connector: BaseConnector,
) -> np.ndarray:
    new_state_vector = state._get_empty()

    index = get_projection_operator_indices(
        state.d,
        state._config.cutoff,
        modes,
        subspace_basis,
    )

    new_state_vector = connector.assign(
        new_state_vector, index, state.state_vector[index]
    )

    return new_state_vector



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq


pytestmark = pytest.mark.benchmark(
    group="fock-boson-sampling",
)


@pytest.fixture
def shots():
    """
    NOTE: SF only supports one shot in the Fock backend.
    """
    return 1


def piquasso_benchmark(
    benchmark, pq_purefock_simulator, example_pq_purefock_state, shots
):
    @benchmark
    def func():
        with pq.Program() as new_program:
            pq.Q(0, 1, 2) | pq.ParticleNumberMeasurement()

        pq_purefock_simulator.execute(
            new_program, shots=shots, initial_state=example_pq_purefock_state
        )


def strawberryfields_benchmark(benchmark, example_sf_fock_state, d, cutoff, shots, sf):
    @benchmark
    def func():
        new_program = sf.Program(d)
        new_engine = sf.Engine(backend="fock", backend_options={"cutoff_dim": cutoff})

        new_program.state = example_sf_fock_state

        with new_program.context as q:
            sf.ops.MeasureFock() | (q[0], q[1], q[2])

        new_engine.run(new_program, shots=shots)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from piquasso import cvqnn


@pytest.mark.parametrize("layer_count, d", [(1, 2), (2, 1), (2, 2), (3, 4), (5, 10)])
def test_generate_random_cvqnn_weights(layer_count, d):
    weights = cvqnn.generate_random_cvqnn_weights(
        layer_count=layer_count, d=d, active_var=0.1, passive_var=0.1
    )

    assert weights.shape[0] == layer_count
    assert weights.shape[1] == 4 * d + 2 * (d * (d - 1) + max(1, d - 1))


def test_create_layers_yields_valid_program():
    d = 3

    weights = cvqnn.generate_random_cvqnn_weights(layer_count=10, d=d)
    layers = cvqnn.create_layers(weights)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.1, phi=np.pi / 5)

        pq.Q() | layers

    simulator = pq.PureFockSimulator(d, config=pq.Config(cutoff=7))

    state = simulator.execute(program).state

    state.validate()


def test_create_program_yields_valid_program():
    d = 3

    weights = cvqnn.generate_random_cvqnn_weights(layer_count=10, d=d)
    program = cvqnn.create_program(weights)

    simulator = pq.PureFockSimulator(d, config=pq.Config(cutoff=7))

    state = simulator.execute(program).state

    state.validate()


def test_create_program_yields_valid_program_for_1_mode():
    d = 1

    weights = cvqnn.generate_random_cvqnn_weights(layer_count=10, d=d)
    program = cvqnn.create_program(weights)

    simulator = pq.PureFockSimulator(d, config=pq.Config(cutoff=7))

    state = simulator.execute(program).state

    state.validate()


def test_create_program_state_vector():
    d = 2

    weights = np.array(
        [
            [
                0.12856391,
                -0.03381019,
                -0.1377453,
                -0.00976438,
                -0.00987302,
                -0.03386338,
                -0.08224322,
                0.14441779,
                -0.00468172,
                -0.00120696,
                -0.04211308,
                -0.19803922,
                -0.0032648,
                0.01448497,
            ],
            [
                0.05684676,
                0.04780328,
                -0.0244914,
                0.01133317,
                0.00669194,
                0.02037788,
                0.03323016,
                0.13712651,
                -0.01981095,
                -0.00937427,
                0.00490652,
                -0.16186161,
                0.01149154,
                0.01250349,
            ],
            [
                0.14877177,
                0.04202048,
                0.16624834,
                0.00349866,
                -0.01133043,
                0.19474302,
                -0.01333464,
                -0.08560477,
                -0.00291201,
                -0.01066722,
                0.03029708,
                -0.06975938,
                0.00673455,
                -0.00111852,
            ],
        ]
    )

    program = cvqnn.create_program(weights)

    simulator = pq.PureFockSimulator(d)

    state = simulator.execute(program).state

    assert np.allclose(
        state.state_vector,
        [
            0.99927124 - 0.00000994j,
            -0.02215289 - 0.00339542j,
            -0.02876671 + 0.00127763j,
            -0.00304377 + 0.0010067j,
            -0.00414854 + 0.0005445j,
            0.00985761 + 0.00054058j,
            0.00013097 - 0.00000967j,
            0.00023877 - 0.00002112j,
            -0.00002112 - 0.00007153j,
            -0.00047556 - 0.00000858j,
        ],
    )


def test_create_program_with_invalid_weights():
    weights = np.empty(shape=(3, 7))

    with pytest.raises(pq.api.exceptions.CVQNNException) as error:
        cvqnn.create_program(weights)

    assert (
        error.value.args[0]
        == f"Invalid number of parameters specified: '{weights.shape[1]}'."
    )


@pytest.mark.parametrize(
    "number_of_parameters, number_of_modes", [(6, 1), (14, 2), (28, 3), (46, 4)]
)
def test_get_number_of_modes(number_of_parameters, number_of_modes):
    assert cvqnn.get_number_of_modes(number_of_parameters) == number_of_modes


def test_get_number_of_modes_with_invalid_number_of_parameters():
    invalid_number_of_parameters = 7

    with pytest.raises(pq.api.exceptions.CVQNNException) as error:
        cvqnn.get_number_of_modes(invalid_number_of_parameters)

    assert (
        error.value.args[0]
        == f"Invalid number of parameters specified: '{invalid_number_of_parameters}'."
    )


def test_get_cvqnn_weight_indices_1_mode():
    weight_indices = cvqnn.get_cvqnn_weight_indices(1)

    expected_weight_indices = [
        np.array([0]),
        np.array([1]),
        np.array([2]),
        np.array([3]),
        np.array([4]),
        np.array([5]),
    ]

    assert all(
        np.allclose(a, b) for a, b in zip(weight_indices, expected_weight_indices)
    )


def test_get_cvqnn_weight_indices():
    weight_indices = cvqnn.get_cvqnn_weight_indices(4)

    expected_weight_indices = [
        np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]),
        np.array([15, 16, 17, 18]),
        np.array([19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33]),
        np.array([34, 35, 36, 37]),
        np.array([38, 39, 40, 41]),
        np.array([42, 43, 44, 45]),
    ]

    assert all(
        np.allclose(a, b) for a, b in zip(weight_indices, expected_weight_indices)
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .passive_linear import passive_linear, beamsplitter5050
from .measurements import (
    post_select_photons,
    imperfect_post_select_photons,
)

from .homodyne import homodyne_measurement

__all__ = [
    "passive_linear",
    "beamsplitter5050",
    "post_select_photons",
    "imperfect_post_select_photons",
    "homodyne_measurement",
]

from typing import Optional, Tuple, Mapping

import random
import numpy as np

from .passive_linear import _apply_passive_linear
from .utils import project_to_subspace

from ...calculations import calculate_state_index_matrix_list

from ..state import PureFockState
from ..batch_state import BatchPureFockState

from piquasso._math.decompositions import euler
from piquasso._math.indices import get_index_in_fock_space
from piquasso._math.fock import (
    get_single_mode_displacement_operator,
    get_creation_operator,
    get_annihilation_operator,
    get_single_mode_squeezing_operator,
    get_single_mode_cubic_phase_operator,
    get_fock_space_basis,
)

from piquasso.instructions import gates

from piquasso.api.result import Result
from piquasso.api.instruction import Instruction
from piquasso.api.connector import BaseConnector
from piquasso.api.exceptions import InvalidState


def particle_number_measurement(
    state: PureFockState, instruction: Instruction, shots: int
) -> Result:
    reduced_state = state.reduced(instruction.modes)

    probability_map = reduced_state.fock_probabilities_map

    samples = random.choices(
        population=list(probability_map.keys()),
        weights=list(probability_map.values()),
        k=shots,
    )

    # NOTE: We choose the last sample for multiple shots.
    sample = samples[-1]

    normalization = _get_normalization(probability_map, sample)

    project_to_subspace(
        state=state,
        subspace_basis=sample,
        modes=instruction.modes,
        normalization=normalization,
        connector=state._connector,
    )

    return Result(state=state, samples=samples)


def vacuum(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    state.reset()

    return Result(state=state)


def _get_normalization(
    probability_map: Mapping[Tuple[int, ...], float], sample: Tuple[int, ...]
) -> float:
    return np.sqrt(1 / probability_map[sample])


def _apply_active_gate_matrix_to_state(
    state_vector: np.ndarray,
    matrix: np.ndarray,
    d: int,
    cutoff: int,
    mode: int,
    connector: BaseConnector,
) -> None:
    @connector.custom_gradient
    def _apply_active_gate_matrix(state_vector, matrix):
        state_vector = connector.preprocess_input_for_custom_gradient(state_vector)
        matrix = connector.preprocess_input_for_custom_gradient(matrix)

        state_index_matrix_list = calculate_state_index_matrix_list(d, cutoff, mode)
        new_state_vector = _calculate_state_vector_after_apply_active_gate(
            state_vector, matrix, state_index_matrix_list, connector
        )
        grad = _create_linear_active_gate_gradient_function(
            state_vector, matrix, state_index_matrix_list, connector
        )
        return new_state_vector, grad

    return _apply_active_gate_matrix(state_vector, matrix)


def _calculate_state_vector_after_apply_active_gate(
    state_vector,
    matrix,
    state_index_matrix_list,
    connector,
):
    np = connector.forward_pass_np

    new_state_vector = np.empty_like(state_vector, dtype=state_vector.dtype)

    is_batch = len(state_vector.shape) == 2

    einsum_string = "ij,jkl->ikl" if is_batch else "ij,jk->ik"

    for state_index_matrix in state_index_matrix_list:
        limit = state_index_matrix.shape[0]
        new_state_vector = connector.assign(
            new_state_vector,
            state_index_matrix,
            np.einsum(
                einsum_string, matrix[:limit, :limit], state_vector[state_index_matrix]
            ),
        )

    return new_state_vector


def _create_linear_active_gate_gradient_function(
    state_vector,
    matrix,
    state_index_matrix_list,
    connector,
):
    def linear_active_gate_gradient(upstream):
        tf = connector._tf
        fallback_np = connector.fallback_np

        static_valued = tf.get_static_value(upstream) is not None

        if static_valued:
            np = connector.fallback_np
            upstream = upstream.numpy()
        else:
            np = connector.np

        cutoff = len(matrix)

        is_batch = len(state_vector.shape) == 2

        matrix_einsum_string = "ijl,kjl->ki" if is_batch else "ij,kj->ki"
        initial_state_einsum_string = "ji,jkl->ikl" if is_batch else "ji,jk->ik"

        reshape_arg = (-1, state_vector.shape[1]) if is_batch else (-1,)

        unordered_gradient_by_initial_state = []
        order_by = []

        gradient_by_matrix = np.zeros(shape=(cutoff, cutoff), dtype=state_vector.dtype)

        conjugated_matrix = np.conj(matrix)
        conjugated_state_vector = np.conj(state_vector)
        for indices in state_index_matrix_list:
            limit = indices.shape[0]

            upstream_slice = upstream[indices]
            state_vector_slice = conjugated_state_vector[indices]

            matrix_slice = conjugated_matrix[:limit, :limit]

            order_by.append(indices.reshape(-1))
            product = np.einsum(
                initial_state_einsum_string, matrix_slice, upstream_slice
            )
            unordered_gradient_by_initial_state.append(product.reshape(*reshape_arg))

            partial_gradient = np.einsum(
                matrix_einsum_string, state_vector_slice, upstream_slice
            )

            if static_valued:
                gradient_by_matrix[:limit, :limit] += partial_gradient
            else:
                gradient_by_matrix += np.pad(
                    partial_gradient,
                    [[0, cutoff - limit], [0, cutoff - limit]],
                    "constant",
                )

        gradient_by_initial_state = np.concatenate(unordered_gradient_by_initial_state)[
            fallback_np.concatenate(order_by).argsort()
        ]

        if static_valued:
            return (
                tf.constant(gradient_by_initial_state),
                tf.constant(gradient_by_matrix),
            )

        return gradient_by_initial_state, gradient_by_matrix

    return linear_active_gate_gradient


def create(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    operator = get_creation_operator(
        instruction.modes, space=space, config=state._config
    )

    state.state_vector = operator @ state.state_vector

    return Result(state=state)


def annihilate(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    operator = get_annihilation_operator(
        instruction.modes, space=space, config=state._config
    )

    state.state_vector = operator @ state.state_vector

    return Result(state=state)


def kerr(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    xi = instruction._all_params["xi"]
    np = state._np

    mode = instruction.modes[0]

    coefficients = np.exp(1j * xi * np.array([basis[mode] ** 2 for basis in space]))

    # NOTE: Transposition is done here in order to work with batch processing.
    state.state_vector = (coefficients * state.state_vector.T).T

    return Result(state=state)


def cross_kerr(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    space = get_fock_space_basis(d=state.d, cutoff=state._config.cutoff)

    np = state._connector.np

    modes = instruction.modes
    xi = instruction._all_params["xi"]
    for index, basis in enumerate(space):
        coefficient = np.exp(1j * xi * basis[modes[0]] * basis[modes[1]])
        state.state_vector = state._connector.assign(
            state.state_vector, index, state.state_vector[index] * coefficient
        )

    return Result(state=state)


def displacement(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    connector = state._connector

    r = instruction._all_params["r"]
    phi = instruction._all_params["phi"]

    wrapped_get_matrix = connector.decorator(get_single_mode_displacement_operator)

    matrix = wrapped_get_matrix(
        r=r,
        phi=phi,
        cutoff=state._config.cutoff,
        complex_dtype=state._config.complex_dtype,
        connector=connector,
    )

    wrapped_apply = connector.decorator(_apply_active_gate_matrix_to_state)

    state.state_vector = wrapped_apply(
        state.state_vector,
        matrix,
        state.d,
        state._config.cutoff,
        instruction.modes[0],
        connector,
    )

    return Result(state=state)


def _apply_squeezing(state, r, phi, mode):
    connector = state._connector

    wrapped_get_matrix = connector.decorator(get_single_mode_squeezing_operator)

    matrix = wrapped_get_matrix(
        r=r,
        phi=phi,
        cutoff=state._config.cutoff,
        complex_dtype=state._config.complex_dtype,
        connector=connector,
    )

    wrapped_apply = connector.decorator(_apply_active_gate_matrix_to_state)

    state.state_vector = wrapped_apply(
        state.state_vector,
        matrix,
        state.d,
        state._config.cutoff,
        mode,
        connector=connector,
    )


def squeezing(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    _apply_squeezing(
        state,
        r=instruction._all_params["r"],
        phi=instruction._all_params["phi"],
        mode=instruction.modes[0],
    )

    return Result(state=state)


def cubic_phase(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    connector = state._connector

    gamma = instruction._all_params["gamma"]

    wrapped_get_matrix = connector.decorator(get_single_mode_cubic_phase_operator)

    matrix = wrapped_get_matrix(
        gamma=gamma,
        cutoff=state._config.cutoff,
        hbar=state._config.hbar,
        connector=connector,
    )

    wrapped_apply = connector.decorator(_apply_active_gate_matrix_to_state)

    state.state_vector = wrapped_apply(
        state.state_vector,
        matrix,
        state.d,
        state._config.cutoff,
        instruction.modes[0],
        connector,
    )

    return Result(state=state)


def linear(
    state: PureFockState, instruction: gates._ActiveLinearGate, shots: int
) -> Result:
    connector = state._connector
    modes = instruction.modes

    np = connector.np

    passive_block = instruction._get_passive_block(state._connector, state._config)
    active_block = instruction._get_active_block(state._connector, state._config)

    symplectic = connector.block(
        [
            [passive_block, active_block],
            [np.conj(active_block), np.conj(passive_block)],
        ],
    )

    unitary_last, squeezings, unitary_first = euler(symplectic, connector)

    _apply_passive_linear(state, unitary_first, modes, connector)

    for mode, r in zip(instruction.modes, squeezings):
        _apply_squeezing(state, r=r, phi=0.0, mode=mode)

    _apply_passive_linear(state, unitary_last, modes, connector)

    return Result(state=state)


def state_vector_instruction(
    state: PureFockState, instruction: Instruction, shots: int
) -> Result:
    if "occupation_numbers" in instruction._all_params:
        occupation_numbers = instruction._all_params["occupation_numbers"]

        if state._config.validate:
            expected_length = len(instruction.modes) if instruction.modes else state.d
            if len(occupation_numbers) != expected_length:
                raise InvalidState(
                    f"The occupation numbers '{occupation_numbers}' are "
                    f"not well-defined on '{expected_length}' modes: "
                    f"instruction={instruction}"
                )

        _add_occupation_number_basis(
            state=state,
            coefficient=instruction._all_params["coefficient"],
            occupation_numbers=occupation_numbers,
            modes=instruction.modes,
        )

    elif "fock_amplitude_map" in instruction._all_params:
        for occupation_numbers, amplitude in instruction._all_params[
            "fock_amplitude_map"
        ].items():
            if state._config.validate:
                expected_length = (
                    len(instruction.modes) if instruction.modes else state.d
                )
                if len(occupation_numbers) != expected_length:
                    raise InvalidState(
                        f"The occupation numbers '{occupation_numbers}' are "
                        f"not well-defined on '{expected_length}' modes: "
                        f"instruction={instruction}"
                    )

            _add_occupation_number_basis(
                state=state,
                coefficient=instruction._all_params["coefficient"] * amplitude,
                occupation_numbers=occupation_numbers,
                modes=instruction.modes,
            )

    return Result(state=state)


def _add_occupation_number_basis(  # type: ignore
    state: PureFockState,
    coefficient: complex,
    occupation_numbers: np.ndarray,
    modes: Optional[Tuple[int, ...]] = None,
) -> None:
    if modes:
        new_occupation_numbers = np.zeros(shape=state.d, dtype=int)

        new_occupation_numbers[modes,] = np.array(occupation_numbers)

        occupation_numbers = new_occupation_numbers

    index = get_index_in_fock_space(occupation_numbers)

    state.state_vector = state._connector.assign(state.state_vector, index, coefficient)


def batch_prepare(state: PureFockState, instruction: Instruction, shots: int) -> Result:
    subprograms = instruction._all_params["subprograms"]
    execute = instruction._all_params["execute"]

    state_vectors = [
        execute(subprogram, shots).state.state_vector for subprogram in subprograms
    ]

    batch_state = BatchPureFockState(
        d=state.d, connector=state._connector, config=state._config
    )

    batch_state._apply_separate_state_vectors(state_vectors)

    return Result(state=batch_state)


def batch_apply(
    state: BatchPureFockState, instruction: Instruction, shots: int
) -> Result:
    subprograms = instruction._all_params["subprograms"]
    execute = instruction._all_params["execute"]

    d = state.d
    connector = state._connector
    config = state._config

    resulting_state_vectors = []

    for state_vector, subprogram in zip(state._batch_state_vectors, subprograms):
        small_state = PureFockState(d=d, connector=connector, config=config)

        small_state.state_vector = state_vector

        resulting_state_vectors.append(
            execute(subprogram, initial_state=small_state).state.state_vector
        )

    state._apply_separate_state_vectors(resulting_state_vectors)

    return Result(state=state)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.instructions.measurements import PostSelectPhotons
from piquasso.api.result import Result

from piquasso._math.fock import get_fock_space_basis
from piquasso._simulators.fock.calculations import get_projection_operator_indices
from piquasso._simulators.fock.general.state import FockState

from ..state import PureFockState


def post_select_photons(
    state: PureFockState, instruction: PostSelectPhotons, shots: int
) -> Result:
    connector = state._connector

    postselect_modes = instruction.params["postselect_modes"]

    photon_counts = instruction.params["photon_counts"]

    index = get_projection_operator_indices(
        d=state.d,
        cutoff=state._config.cutoff,
        modes=postselect_modes,
        basis_vector=photon_counts,
    )
    small_index = connector.fallback_np.arange(index.shape[0])

    new_state = PureFockState(
        d=state.d - len(postselect_modes),
        connector=state._connector,
        config=state._config,
    )

    new_state.state_vector = connector.assign(
        new_state.state_vector, small_index, state.state_vector[index]
    )

    return Result(state=new_state)


def imperfect_post_select_photons(
    state: PureFockState, instruction: PostSelectPhotons, shots: int
) -> Result:
    np = state._connector.np

    postselect_modes = instruction.params["postselect_modes"]

    photon_counts = instruction.params["photon_counts"]

    detector_efficiency_matrix = instruction.params["detector_efficiency_matrix"]

    postselect_basis = get_fock_space_basis(
        d=len(postselect_modes), cutoff=len(detector_efficiency_matrix)
    )

    new_state = FockState(
        d=state.d - len(postselect_modes),
        connector=state._connector,
        config=state._config,
    )

    for occupation_numbers in postselect_basis:
        detector_probability = np.prod(
            detector_efficiency_matrix[(photon_counts, occupation_numbers)]
        )

        index = get_projection_operator_indices(
            state.d, state._config.cutoff, postselect_modes, occupation_numbers
        )

        small_index = np.arange(index.shape[0])

        state_vector = np.zeros(
            shape=new_state.density_matrix.shape[0],
            dtype=new_state.density_matrix.dtype,
        )

        state_vector = state._connector.assign(
            state_vector, small_index, state.state_vector[index]
        )

        new_state._density_matrix += detector_probability * np.outer(
            state_vector, np.conj(state_vector)
        )

    return Result(state=new_state)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq

import numpy as np

from itertools import chain, combinations


from piquasso._math.torontonian import torontonian, loop_torontonian
from piquasso._math.transformations import xpxp_to_xxpp_indices


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(iterable, r) for r in range(len(s) + 1))


def torontonian_naive(A: np.ndarray) -> complex:
    """
    Naive torontonian implementation from formulas provided in
    https://arxiv.org/abs/2202.04600.
    """
    d = A.shape[0] // 2

    indices = xpxp_to_xxpp_indices(d)

    A = A[np.ix_(indices, indices)]

    if d == 0:
        return 1.0 + 0j

    ret = 0.0 + 0j

    for subset in powerset(range(0, d)):
        index = np.ix_(subset, subset)

        A_reduced = np.block(
            [
                [A[:d, :d][index], A[:d, d:][index]],
                [A[d:, :d][index], A[d:, d:][index]],
            ]
        )

        factor = 1.0 if ((d - len(subset)) % 2 == 0) else -1.0

        inner_mat = np.identity(len(A_reduced)) - A_reduced

        determinant = np.linalg.det(inner_mat)

        summand = factor / np.sqrt(determinant.real + 0.0j)

        ret += summand

    return ret


def loop_torontonian_naive(A, gamma):
    """
    Naive loop torontonian implementation from formulas provided in
    https://arxiv.org/abs/2202.04600.
    """

    d = A.shape[0] // 2

    indices = xpxp_to_xxpp_indices(d)

    A = A[np.ix_(indices, indices)]
    gamma = gamma[indices]

    if d == 0:
        return 1.0

    ret = 0.0

    for subset in powerset(range(0, d)):
        subset = np.array(subset, dtype=int)
        index = np.ix_(subset, subset)

        vector_index = np.concatenate([subset, subset + d])

        gamma_reduced = gamma[vector_index]

        A_reduced = np.block(
            [
                [A[:d, :d][index], A[:d, d:][index]],
                [A[d:, :d][index], A[d:, d:][index]],
            ]
        )

        inner_mat = np.identity(len(A_reduced)) - A_reduced

        exponential_term = np.exp(
            gamma_reduced @ np.linalg.inv(inner_mat) @ gamma_reduced / 2
        )

        factor = 1.0 if ((d - len(subset)) % 2 == 0) else -1.0

        determinant = np.linalg.det(inner_mat)

        summand = factor * exponential_term / np.sqrt(determinant)

        ret += summand

    return ret


def test_torontonian_empty():
    matrix = np.array([[]], dtype=float)

    assert np.isclose(torontonian(matrix), 1.0)


def test_torontonian_2_by_2_float32():
    matrix = np.array(
        [
            [0.6, 0.4],
            [0.4, 0.5],
        ],
        dtype=np.float32,
    )

    output = torontonian(matrix)

    assert output.dtype == np.float32

    assert np.isclose(output, torontonian_naive(matrix))
    assert np.isclose(output, 4.0)


def test_torontonian_2_by_2_float64():
    matrix = np.array(
        [
            [0.6, 0.4],
            [0.4, 0.5],
        ],
        dtype=np.float64,
    )

    output = torontonian(matrix)

    assert output.dtype == np.float64

    assert np.isclose(output, torontonian_naive(matrix))
    assert np.isclose(output, 4.0)


@pytest.mark.monkey
def test_torontonian_4_by_4_random():
    A = np.random.rand(4, 4)
    matrix = A @ A.T

    matrix /= max(np.linalg.eigvals(matrix)) + 1.0

    torontonian(matrix)

    assert np.isclose(torontonian(matrix), torontonian_naive(matrix))


@pytest.mark.monkey
def test_torontonian_6_by_6_random():
    A = np.random.rand(6, 6)
    matrix = A @ A.T

    matrix /= max(np.linalg.eigvals(matrix)) + 1.0

    torontonian(matrix)

    assert np.isclose(torontonian(matrix), torontonian_naive(matrix))


@pytest.mark.monkey
@pytest.mark.parametrize("d", (1, 2, 3, 4, 5, 6))
def test_loop_torontonian_random(d, generate_unitary_matrix):
    simulator = pq.GaussianSimulator(d=d)

    program = pq.Program(
        instructions=[pq.Vacuum()]
        + [pq.Displacement(r=np.random.rand()).on_modes(i) for i in range(d)]
        + [pq.Squeezing(r=np.random.rand()).on_modes(i) for i in range(d)]
        + [pq.Interferometer(generate_unitary_matrix(d))]
    )

    state = simulator.execute(program).state

    xpxp_covariance_matrix = state.xpxp_covariance_matrix

    sigma: np.ndarray = (xpxp_covariance_matrix / 2 + np.identity(2 * d)) / 2

    input_matrix = np.identity(len(sigma), dtype=float) - np.linalg.inv(sigma)
    displacement_vector = state.xpxp_mean_vector

    assert np.isclose(
        loop_torontonian(input_matrix, displacement_vector),
        loop_torontonian_naive(input_matrix, displacement_vector),
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq


def test_StateVector_is_valid_specifying_floats_close_to_integers():
    pq.StateVector([1, 1.0, 2.0, 0.0])


def test_StateVector_raises_InvalidState_when_nonintegers_specified():
    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        pq.StateVector([1, 1.3, 2.4])

    assert error.value.args[0] == (
        "Invalid occupation numbers: occupation_numbers=[1, 1.3, 2.4]\n"
        "Occupation numbers must contain non-negative integers."
    )


def test_StateVector_stores_fock_amplitude_map():
    amplitude_map = {(0,): 1.0, (1,): 0.0}

    preparation = pq.StateVector(fock_amplitude_map=amplitude_map)

    assert preparation.params["fock_amplitude_map"] == amplitude_map


def test_StateVector_raises_InvalidParameter_when_no_arguments():
    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        pq.StateVector()

    assert error.value.args[0] == (
        "Either 'occupation_numbers' or 'fock_amplitude_map' must be provided."
    )


def test_StateVector_raises_InvalidParameter_when_both_arguments_given():
    amplitude_map = {(0,): 1.0}

    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        pq.StateVector([0], fock_amplitude_map=amplitude_map)

    assert error.value.args[0] == (
        "Only one of 'occupation_numbers' or 'fock_amplitude_map' can be provided."
    )


def test_StateVector_raises_InvalidState_when_fock_amplitude_map_keys_invalid():
    amplitude_map = {(0, 1.2): 1.0}

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        pq.StateVector(fock_amplitude_map=amplitude_map)

    assert error.value.args[0] == (
        "Invalid occupation numbers in fock_amplitude_map: {(0, 1.2): 1.0}\n"
        "Occupation numbers must contain non-negative integers."
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

import time
import json

import strawberryfields as sf

import tensorflow as tf

import piquasso as pq

from piquasso import cvqnn

tf.get_logger().setLevel("ERROR")
np.set_printoptions(suppress=True, linewidth=200)

RUN = "pq"
GRAPH_COMPILE = True

n_layers = 4
n_mode_range = list(range(2, 15 + 1))
cutoff = 10
np_seed = 42

timestamp = int(time.time())
filename = (
    f"./result_{RUN}_graph={GRAPH_COMPILE}_"
    f"d={min(n_mode_range)}-{max(n_mode_range)}_l={n_layers}_c={cutoff}_{timestamp}"
    ".json"
)

decorator = tf.function if GRAPH_COMPILE else lambda x: x


def main():
    # Set TensorFlow to use only the CPU
    tf.config.set_visible_devices([], "GPU")

    # Verify devices
    # This step is optional, just to confirm that GPU is not being used.
    physical_devices = tf.config.list_physical_devices()
    print("Available devices:", physical_devices)

    # Create a process to run the benchmark function with queue, pq, and sf as arguments
    result = {"RUN": RUN, "cutoff": cutoff, "np_seed": np_seed, "runtimes": {}}
    with tf.device("/CPU:0"):
        for d in n_mode_range:
            partial_result = benchmark(
                n_layers=n_layers,
                d=d,
                cutoff=cutoff,
                np_seed=np_seed,
            )
            result["runtimes"][d] = partial_result

            with open(filename, "w") as json_file:
                json.dump(result, json_file, indent=4)


def benchmark(n_layers, d, cutoff, np_seed):
    weights = make_weights(layer_count=n_layers, d=d, np_seed=np_seed)

    runtimes = []

    if RUN == "pq":
        func_to_benchmark = _calculate_piquasso_results
    else:
        func_to_benchmark = _calculate_strawberryfields_results

    # Warmup
    start_time = time.time()
    func_to_benchmark(weights, cutoff)
    print("Warmup time:", time.time() - start_time)

    runs = 100
    k = 0
    while k < runs:
        print(".", end="", flush=True)
        weights = make_weights(layer_count=n_layers, d=d, np_seed=42 + k + 1)

        now = time.time()
        _, _, grad = func_to_benchmark(weights, cutoff)
        runtime = time.time() - now
        runtimes.append(runtime)

        assert grad is not None, "'grad is None', graph might be disconnected"

        if runtime > 1.0:
            runs = 10

        k = k + 1

    print()

    return runtimes


def make_weights(layer_count, d, np_seed):
    active_sd = 0.1
    passive_sd = 1.0
    np.random.seed(np_seed)

    M = int(d * (d - 1)) + max(1, d - 1)

    int1_weights = tf.Variable(
        np.random.normal(size=(layer_count, M), loc=0.0, scale=passive_sd)
    )
    s_weights = tf.Variable(
        np.random.normal(size=(layer_count, d), loc=0.0, scale=active_sd)
    )
    int2_weights = tf.Variable(
        np.random.normal(size=(layer_count, M), loc=0.0, scale=passive_sd)
    )
    dr_weights = tf.Variable(
        np.random.normal(size=(layer_count, d), loc=0.0, scale=active_sd)
    )
    dp_weights = tf.Variable(
        np.random.normal(size=(layer_count, d), loc=0.0, scale=passive_sd)
    )
    k_weights = tf.Variable(
        np.random.normal(size=(layer_count, d), loc=0.0, scale=active_sd)
    )

    weights = tf.cast(
        tf.concat(
            [int1_weights, s_weights, int2_weights, dr_weights, dp_weights, k_weights],
            axis=1,
        ),
        dtype=tf.float64,
    )

    weights = tf.Variable(weights)

    return weights


def _pq_state_vector(weights, cutoff):
    d = cvqnn.get_number_of_modes(weights.shape[1])

    simulator = pq.PureFockSimulator(
        d=d,
        config=pq.Config(cutoff=cutoff),
        connector=pq.TensorflowConnector(
            decorate_with=(tf.function() if GRAPH_COMPILE else None)
        ),
    )

    program = cvqnn.create_program(weights)

    state = simulator.execute(program).state

    return state.state_vector


@decorator
def _calculate_piquasso_results(weights, cutoff):
    with tf.GradientTape() as tape:
        state_vector = _pq_state_vector(weights, cutoff)

        target_state_vector = tf.ones(
            shape=state_vector.shape, dtype=state_vector.dtype
        ) / tf.math.sqrt(tf.cast(len(state_vector), state_vector.dtype))

        loss = tf.math.reduce_mean(tf.abs(target_state_vector - state_vector) ** 2)

    return state_vector, loss, tape.gradient(loss, weights)


def _calculate_strawberryfields_results(weights, cutoff):
    layer_count = weights.shape[0]
    d = cvqnn.get_number_of_modes(weights.shape[1])

    eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff})
    qnn = sf.Program(d)

    num_params = np.prod(weights.shape)

    sf_params = np.arange(num_params).reshape(weights.shape).astype(str)
    sf_params = np.array([qnn.params(*i) for i in sf_params])

    with qnn.context as q:
        for k in range(layer_count):
            _sf_layer(sf_params[k], q)

    with tf.GradientTape() as tape:
        mapping = {
            p.name: w for p, w in zip(sf_params.flatten(), tf.reshape(weights, [-1]))
        }

        state = eng.run(qnn, args=mapping).state
        state_vector = state.ket()
        target_state_vector = tf.ones(
            shape=state_vector.shape, dtype=state_vector.dtype
        ) / tf.math.sqrt(tf.cast(len(state_vector), state_vector.dtype))

        loss = tf.math.reduce_mean(tf.abs(target_state_vector - state_vector) ** 2)

    return state_vector, loss, tape.gradient(loss, weights)


def _sf_interferometer(params, q):
    N = len(q)
    theta = params[: N * (N - 1) // 2]
    phi = params[N * (N - 1) // 2 : N * (N - 1)]
    rphi = params[-N + 1 :]

    if N == 1:
        sf.ops.Rgate(rphi[0]) | q[0]
        return

    n = 0

    for j in range(N):
        for k, (q1, q2) in enumerate(zip(q[:-1], q[1:])):
            if (j + k) % 2 != 1:
                sf.ops.BSgate(theta[n], phi[n]) | (q1, q2)
                n += 1

    for i in range(max(1, N - 1)):
        sf.ops.Rgate(rphi[i]) | q[i]


def _sf_layer(params, q):
    N = len(q)
    M = int(N * (N - 1)) + max(1, N - 1)

    int1 = params[:M]
    s = params[M : M + N]
    int2 = params[M + N : 2 * M + N]
    dr = params[2 * M + N : 2 * M + 2 * N]
    dp = params[2 * M + 2 * N : 2 * M + 3 * N]
    k = params[2 * M + 3 * N : 2 * M + 4 * N]

    _sf_interferometer(int1, q)

    for i in range(N):
        sf.ops.Sgate(s[i]) | q[i]

    _sf_interferometer(int2, q)

    for i in range(N):
        sf.ops.Dgate(dr[i], dp[i]) | q[i]
        sf.ops.Kgate(k[i]) | q[i]


if __name__ == "__main__":
    main()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numba as nb
import numpy as np


@nb.njit(cache=True)
def _get_reflection_vector(input):
    size = len(input)

    reflect_vector = np.empty(size, dtype=input.dtype)
    norm_v_sqr = 0.0
    for i in range(0, size):
        element = input[i]
        norm_v_sqr += np.real(element) ** 2 + np.imag(element) ** 2
        reflect_vector[i] = element

    sigma = np.sqrt(norm_v_sqr)

    abs_val = np.abs(reflect_vector[0])
    norm_v_sqr = 2 * (norm_v_sqr + abs_val * sigma)

    if abs_val != 0:
        addend = reflect_vector[0] / abs_val * sigma
        reflect_vector[0] += addend
    else:
        reflect_vector[0] += sigma

    if norm_v_sqr == 0.0:
        return reflect_vector, 0.0

    norm_v = np.sqrt(norm_v_sqr)

    reflect_vector /= norm_v

    return reflect_vector, 1.0


HOUSEHOLDER_CUTOFF = 40


@nb.njit(cache=True)
def _mult_a_bconj(a, b):
    return a * np.conj(b)


@nb.njit(cache=True)
def _calc_vH_times_A(A, v, vH_times_A):
    if A.shape[1] > HOUSEHOLDER_CUTOFF:

        cols_mid = A.shape[1] // 2
        A1 = A[:, :cols_mid]
        vH_times_A_1 = vH_times_A[:cols_mid]
        _calc_vH_times_A(A1, v, vH_times_A_1)

        A2 = A[:, cols_mid:]
        vH_times_A_2 = vH_times_A[cols_mid:]

        _calc_vH_times_A(A2, v, vH_times_A_2)

    elif A.shape[0] > HOUSEHOLDER_CUTOFF:

        rows_mid = A.shape[0] // 2
        A1 = A[:rows_mid]
        v1 = v[:rows_mid]

        _calc_vH_times_A(A1, v1, vH_times_A)

        A2 = A[rows_mid:]
        v2 = v[rows_mid:]
        _calc_vH_times_A(A2, v2, vH_times_A)
    else:
        sizeH = len(v)

        for row_idx in range(0, sizeH - 1, 2):
            data_A = A[row_idx]
            data_A2 = A[row_idx + 1]

            for j in range(0, A.shape[1] - 1, 2):
                vH_times_A[j] = vH_times_A[j] + _mult_a_bconj(data_A[j], v[row_idx])
                vH_times_A[j + 1] = vH_times_A[j + 1] + _mult_a_bconj(
                    data_A[j + 1], v[row_idx]
                )
                vH_times_A[j] = vH_times_A[j] + _mult_a_bconj(
                    data_A2[j], v[row_idx + 1]
                )
                vH_times_A[j + 1] = vH_times_A[j + 1] + _mult_a_bconj(
                    data_A2[j + 1], v[row_idx + 1]
                )

            if A.shape[1] % 2 == 1:
                j = A.shape[1] - 1
                vH_times_A[j] = vH_times_A[j] + _mult_a_bconj(data_A[j], v[row_idx])
                vH_times_A[j] = vH_times_A[j] + _mult_a_bconj(
                    data_A2[j], v[row_idx + 1]
                )

        if sizeH % 2 == 1:
            row_idx = sizeH - 1

            data_A = A[row_idx]

            for j in range(0, A.shape[1] - 1, 2):
                vH_times_A[j] = vH_times_A[j] + _mult_a_bconj(data_A[j], v[row_idx])
                vH_times_A[j + 1] = vH_times_A[j + 1] + _mult_a_bconj(
                    data_A[j + 1], v[row_idx]
                )

            if A.shape[1] % 2 == 1:
                j = A.shape[1] - 1
                vH_times_A[j] = vH_times_A[j] + _mult_a_bconj(data_A[j], v[row_idx])


@nb.njit(cache=True)
def _calc_vov_times_A(A, v, vH_times_A):
    if A.shape[1] > HOUSEHOLDER_CUTOFF:

        cols_mid = A.shape[1] // 2
        A1 = A[:, :cols_mid]
        vH_times_A_1 = vH_times_A[:cols_mid]

        _calc_vov_times_A(A1, v, vH_times_A_1)

        A2 = A[:, cols_mid:]
        vH_times_A_2 = vH_times_A[cols_mid:]

        _calc_vov_times_A(A2, v, vH_times_A_2)

    elif A.shape[0] > HOUSEHOLDER_CUTOFF:
        rows_mid = A.shape[0] // 2
        A1 = A[:rows_mid]
        v1 = v[:rows_mid]

        _calc_vov_times_A(A1, v1, vH_times_A)

        A2 = A[rows_mid:]
        v2 = v[rows_mid:]
        _calc_vov_times_A(A2, v2, vH_times_A)

    else:
        size_v = len(v)

        for row_idx in range(0, size_v - 1, 2):
            data_A = A[row_idx]
            data_A2 = A[row_idx + 1]

            factor = v[row_idx] * 2.0
            factor2 = v[row_idx + 1] * 2.0

            for kdx in range(0, A.shape[1] - 1, 2):
                data_A[kdx] = data_A[kdx] - factor * vH_times_A[kdx]
                data_A2[kdx] = data_A2[kdx] - factor2 * vH_times_A[kdx]
                data_A[kdx + 1] = data_A[kdx + 1] - factor * vH_times_A[kdx + 1]
                data_A2[kdx + 1] = data_A2[kdx + 1] - factor2 * vH_times_A[kdx + 1]

            if A.shape[1] % 2 == 1:
                kdx = A.shape[1] - 1
                data_A[kdx] = data_A[kdx] - factor * vH_times_A[kdx]
                data_A2[kdx] = data_A2[kdx] - factor2 * vH_times_A[kdx]

        if size_v % 2 == 1:

            row_idx = len(v) - 1

            data_A = A[row_idx]

            factor = v[row_idx] * 2.0

            for kdx in range(0, A.shape[1] - 1, 2):
                data_A[kdx] = data_A[kdx] - factor * vH_times_A[kdx]
                data_A[kdx + 1] = data_A[kdx + 1] - factor * vH_times_A[kdx + 1]

            if A.shape[1] % 2 == 1:
                kdx = A.shape[1] - 1
                data_A[kdx] = data_A[kdx] - factor * vH_times_A[kdx]


@nb.njit(cache=True)
def _apply_householder_rows(A, v):
    """
    Apply householder transformation on a matrix `A' = (1 - 2*v o v/v^2) A` for one
    specific reflection vector `v`.
    """
    vH_times_A = np.zeros(shape=(A.shape[1]), dtype=A.dtype)

    _calc_vH_times_A(A, v, vH_times_A)

    _calc_vov_times_A(A, v, vH_times_A)


@nb.njit(cache=True)
def _apply_householder_cols_req(A, v):
    """
    Apply householder transformation on a matrix `A' = A(1 - 2*v o v)` for one
    specific reflection vector `v`.
    """
    sizeH = len(v)

    for idx in range(0, A.shape[0] - 1, 2):
        data_A = A[idx]
        data_A2 = A[idx + 1]

        factor = 0.0
        factor2 = 0.0
        for v_idx in range(0, sizeH):
            factor = factor + data_A[v_idx] * v[v_idx]
            factor2 = factor2 + data_A2[v_idx] * v[v_idx]

        factor = factor * 2.0
        factor2 = factor2 * 2.0
        for jdx in range(0, sizeH):
            data_A[jdx] = data_A[jdx] - _mult_a_bconj(factor, v[jdx])
            data_A2[jdx] = data_A2[jdx] - _mult_a_bconj(factor2, v[jdx])

    if A.shape[0] % 2 == 1:
        data_A = A[-1]

        factor = 0.0
        for v_idx in range(0, sizeH):
            factor = factor + data_A[v_idx] * v[v_idx]

        factor = factor * 2.0
        for jdx in range(0, sizeH):
            data_A[jdx] = data_A[jdx] - _mult_a_bconj(factor, v[jdx])


@nb.njit(cache=True)
def transform_matrix_to_hessenberg(mtx):
    """Reduce a general matrix to upper Hessenberg form."""
    for idx in range(1, len(mtx) - 1):
        reflect_vector, norm_v_sqr = _get_reflection_vector(mtx[idx:, idx - 1])

        if norm_v_sqr == 0.0:
            continue

        mtx_strided = mtx[idx:, (idx - 1) :]
        _apply_householder_rows(mtx_strided, reflect_vector)

        mtx_strided = mtx[:, idx:]
        _apply_householder_cols_req(mtx_strided, reflect_vector)


@nb.njit(cache=True)
def transform_matrix_to_hessenberg_loop(mtx, Lv, Rv):
    for idx in range(1, len(mtx) - 1):
        reflect_vector, norm_v_sqr = _get_reflection_vector(mtx[idx:, idx - 1])

        if norm_v_sqr == 0.0:
            continue

        mtx_strided = mtx[idx:, (idx - 1) :]
        _apply_householder_rows(mtx_strided, reflect_vector)

        Lv_strided = Lv[None, idx:]
        _apply_householder_cols_req(Lv_strided, reflect_vector)

        mtx_strided = mtx[:, idx:]
        _apply_householder_cols_req(mtx_strided, reflect_vector)

        Rv_strided = Rv[idx:, None]
        _apply_householder_rows(Rv_strided, reflect_vector)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from functools import partial

from ..connector import BuiltinConnector


class JaxConnector(BuiltinConnector):
    """The calculations for a simulation using JAX.

    Example usage::

        import numpy as np
        import piquasso as pq
        from jax import jit, grad

        def example_func(r, theta):
            jax_connector = pq.JaxConnector()

            simulator = pq.PureFockSimulator(
                d=2,
                config=pq.Config(cutoff=5, dtype=np.float32, normalize=False),
                connector=jax_connector,
            )

            with pq.Program() as program:
                pq.Q() | pq.Vacuum()

                pq.Q(0) | pq.Displacement(r=r)

                pq.Q(0, 1) | pq.Beamsplitter(theta)

            state = simulator.execute(program).state

            return state.fock_probabilities[0]

        compiled_func = jit(example_func)

        vacuum_probability = compiled_func(0.1, np.pi / 7)

        compiled_grad_func = jit(grad(compiled_func))

        vacuum_probability_grad = compiled_grad_func(0.2, np.pi / 11)

    Note:
        This feature is still experimental.

    Note:
        Currently JIT compilation only works with the config variable `normalize=False`.

    Note:
        Only CPU calculations are supported currently.
    """

    # JAX calculations are often JIT-compiled, while JIT-compilation and conditionals
    # don't work together, when the condition depends on a tracer. Therefore,
    # conditionals must be disabled in this case
    allow_conditionals = False

    def __init__(self):
        try:
            import jax.numpy as jnp
            import jax
        except ImportError:
            raise ImportError(
                "You have invoked a feature which requires 'jax'.\n"
                "You can install JAX via:\n"
                "\n"
                "pip install piquasso[jax]"
            )

        # NOTE: Because lots of calculations are still done using NumPy, and NumPy
        # prefers double precision, Piquasso uses double precision too, therefore it is
        # better to enable this config variable. Theoretically, one could set
        # `pq.Config(dtype=np.float32)`, but this might not always work.
        from jax import config

        config.update("jax_enable_x64", True)

        self.np = jnp
        self._scipy = jax.scipy
        self._jax = jax
        self.fallback_np = np
        self.forward_pass_np = jnp

    def preprocess_input_for_custom_gradient(self, value):
        return value

    def assign(self, array, index, value):
        return array.at[index].set(value)

    def scatter(self, indices, updates, shape):
        embedded_matrix = self.np.zeros(shape, dtype=updates[0].dtype)
        indices_array = self.np.array(indices)
        composite_index = tuple([indices_array[:, i] for i in range(len(shape))])

        return embedded_matrix.at[composite_index].set(self.np.array(updates))

    def polar(self, a, side):
        # NOTE: The default QDWH algorithm does not support left polar decomposition
        # in `jax.scipy.linalg.polar`, so we have to switch to SVD.
        return self._scipy.linalg.polar(a, side, method="svd")

    def block_diag(self, *args, **kwargs):
        return self._scipy.linalg.block_diag(*args, **kwargs)

    def block(self, *args, **kwargs):
        return self.np.block(*args, **kwargs)

    def logm(self, *args, **kwargs):
        return partial(self._scipy.linalg.funm, func=self.np.log)(*args, **kwargs)

    def expm(self, *args, **kwargs):
        return self._scipy.linalg.expm(*args, **kwargs)

    def powm(self, *args, **kwargs):
        return self.np.linalg.matrix_power(*args, **kwargs)

    def sqrtm(self, *args, **kwargs):
        return self._scipy.linalg.sqrtm(*args, **kwargs)

    def svd(self, *args, **kwargs):
        return self.np.linalg.svd(*args, **kwargs)

    def schur(self, *args, **kwargs):
        return self._scipy.linalg.schur(*args, **kwargs)

    def permanent(self, *args, **kwargs):
        from piquasso._math.jax.permanent import permanent_with_reduction

        return permanent_with_reduction(*args, **kwargs)

    def hafnian(self, matrix, reduce_on):
        raise NotImplementedError()

    def loop_hafnian(self, matrix, diagonal, reduce_on):
        from piquasso._math.jax.hafnian import loop_hafnian_with_reduction

        return loop_hafnian_with_reduction(matrix, diagonal, reduce_on)

    def loop_hafnian_batch(self, matrix, diagonal, reduce_on, cutoff):
        raise NotImplementedError()

    def calculate_interferometer_on_fermionic_fock_space(self, matrix, cutoff):
        from .connections import (
            calculate_interferometer_on_fermionic_fock_space,
        )

        return calculate_interferometer_on_fermionic_fock_space(matrix, cutoff)

    def apply_fermionic_passive_linear_to_state_vector(
        self, representations, state_vector, modes, d, cutoff
    ):
        from .connections import (
            apply_fermionic_passive_linear_to_state_vector,
        )

        return apply_fermionic_passive_linear_to_state_vector(
            representations, state_vector, modes, d, cutoff
        )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def test_example_programs_result_in_the_same_state(
    example_pq_gaussian_state,
    example_sf_gaussian_state,
):
    # NOTE: While in SF they use the xp-ordered mean and covariance by default,
    # we access it by the `xxpp_` prefixes.
    assert np.allclose(
        example_pq_gaussian_state.xxpp_mean_vector,
        example_sf_gaussian_state.means(),
    )

    # NOTE: We use a different definition for the covariance in piquasso, that is the
    # reason for the scaling by 2.
    assert np.allclose(
        example_pq_gaussian_state.xxpp_covariance_matrix / 2,
        example_sf_gaussian_state.cov(),
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Any, Generator, Dict

import numpy as np
from piquasso.api.connector import BaseConnector

from piquasso.api.config import Config
from piquasso.api.exceptions import InvalidState, PiquassoException
from piquasso._math.linalg import is_selfadjoint
from piquasso._math.fock import cutoff_fock_space_dim, get_fock_space_basis
from piquasso._math.indices import (
    get_index_in_fock_space,
    get_index_in_fock_space_array,
)

from ..state import BaseFockState


class FockState(BaseFockState):
    """Object to represent a general bosonic state in the Fock basis.

    Note:
        If you only work with pure states, it is advised to use
        :class:`~piquasso._simulators.fock.pure.state.PureFockState` instead.
    """

    def __init__(
        self, *, d: int, connector: BaseConnector, config: Optional[Config] = None
    ) -> None:
        """
        Args:
            d (int): The number of modes.
            connector (BaseConnector): Instance containing calculation functions.
            config (Config): Instance containing constants for the simulation.
        """

        super().__init__(d=d, connector=connector, config=config)

        self._density_matrix = self._get_empty()

    def _get_empty(self) -> np.ndarray:
        state_vector_size = cutoff_fock_space_dim(cutoff=self._config.cutoff, d=self.d)
        return np.zeros(
            shape=(state_vector_size,) * 2, dtype=self._config.complex_dtype
        )

    def reset(self) -> None:
        self._density_matrix = self._get_empty()
        self._density_matrix[0, 0] = 1.0

    def _as_mixed(self):
        return self.copy()

    @classmethod
    def from_fock_state(cls, state: BaseFockState) -> "FockState":
        """Instantiation using another :class:`BaseFockState` instance.

        Args:
            state (BaseFockState):
                The instance from which a :class:`FockState` instance is created.
        """

        new_instance = cls(d=state.d, connector=state._connector, config=state._config)

        new_instance._density_matrix = state.density_matrix

        return new_instance

    @property
    def nonzero_elements(
        self,
    ) -> Generator[Tuple[complex, Tuple], Any, None]:
        np = self._connector.np

        nonzero_indices = np.nonzero(self._density_matrix)

        row_indices, col_indices = nonzero_indices

        row_occupation_numbers = self._space[row_indices]
        col_occupation_numbers = self._space[col_indices]

        nonzero_elements = self._density_matrix[nonzero_indices]

        for index, coefficient in enumerate(nonzero_elements):
            yield coefficient, (
                tuple(row_occupation_numbers[index]),
                tuple(col_occupation_numbers[index]),
            )

    def __str__(self) -> str:
        return " + ".join(
            [
                str(coefficient) + str(basis)
                for coefficient, basis in self.nonzero_elements
            ]
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FockState):
            return False
        return np.allclose(self._density_matrix, other._density_matrix)

    @property
    def density_matrix(self) -> np.ndarray:
        cardinality = cutoff_fock_space_dim(d=self.d, cutoff=self._config.cutoff)

        return self._density_matrix[:cardinality, :cardinality]

    def get_particle_detection_probability(
        self, occupation_number: np.ndarray
    ) -> float:
        if self._config.validate and len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        index = get_index_in_fock_space(occupation_number)
        return np.diag(self._density_matrix)[index].real

    @property
    def fock_probabilities(self) -> np.ndarray:
        return self._connector.np.real(self._connector.np.diag(self._density_matrix))

    @property
    def fock_probabilities_map(self) -> Dict[Tuple[int, ...], float]:
        probability_map: Dict[Tuple[int, ...], float] = {}

        for index, basis in enumerate(self._space):
            probability_map[tuple(basis)] = np.abs(self._density_matrix[index, index])

        return probability_map

    def reduced(self, modes: Tuple[int, ...]) -> "FockState":
        np = self._connector.np
        fallback_np = self._connector.fallback_np
        d = self.d

        if modes == tuple(range(d)):
            return self

        outer_modes = self._get_auxiliary_modes(modes)

        inner_size = len(modes)
        outer_size = d - inner_size

        reduced_state = FockState(
            d=inner_size, connector=self._connector, config=self._config
        )

        density_list = [
            [0.0 for _ in range(reduced_state._density_matrix.shape[1])]
            for _ in range(reduced_state._density_matrix.shape[0])
        ]

        cutoff = self._config.cutoff
        inner_space = get_fock_space_basis(d=inner_size, cutoff=cutoff)
        outer_space = get_fock_space_basis(d=outer_size, cutoff=cutoff)

        index_list = []

        basis_matrix = fallback_np.empty(
            shape=(outer_space.shape[0], self.d), dtype=int
        )

        for inner_basis in inner_space:
            size = cutoff_fock_space_dim(
                cutoff=cutoff - fallback_np.sum(inner_basis), d=outer_size
            )

            basis_matrix[:size, modes] = inner_basis

            basis_matrix[:size, outer_modes] = outer_space[:size]

            partial_index_list = get_index_in_fock_space_array(basis_matrix[:size])

            index_list.append(partial_index_list)

        for i, ix1 in enumerate(index_list):
            for j, ix2 in enumerate(index_list):
                minlen = min(len(ix1), len(ix2))
                density_list[i][j] = np.sum(
                    self._density_matrix[ix1[:minlen], ix2[:minlen]]
                )

        reduced_state._density_matrix = np.array(density_list)

        return reduced_state

    @property
    def norm(self) -> float:
        np = self._connector.np
        return np.real(np.trace(self._density_matrix))

    def normalize(self) -> None:
        """Normalizes the density matrix to have a trace of 1.

        Raises:
            RuntimeError: Raised if the current norm of the state is too close to 0.
        """
        norm = self.norm

        if self._config.validate and np.isclose(norm, 0):
            raise InvalidState("The norm of the state is 0.")

        self._density_matrix = self._density_matrix / self.norm

    def validate(self) -> None:
        """Validates the represented state.

        Raises:
            InvalidState:
                Raised, if the density matrix is not positive semidefinite, not
                self-adjoint or the trace of the density matrix is not 1.
        """
        if not self._config.validate:
            return

        if not is_selfadjoint(self._density_matrix):
            raise InvalidState(
                "The density matrix is not self-adjoint:\n"
                f"density_matrix={self._density_matrix}"
            )

        fock_probabilities = self.fock_probabilities

        if not np.all(fock_probabilities >= 0.0):
            raise InvalidState(
                "The density matrix is not positive semidefinite.\n"
                f"fock_probabilities={fock_probabilities}"
            )

        trace_of_density_matrix = sum(fock_probabilities)

        if not np.isclose(trace_of_density_matrix, 1.0):
            raise InvalidState(
                f"The trace of the density matrix is {trace_of_density_matrix}, "
                "instead of 1.0:\n"
                f"fock_probabilities={fock_probabilities}"
            )

    def quadratures_mean_variance(
        self, modes: Tuple[int, ...], phi: float = 0
    ) -> Tuple[float, float]:
        r"""This method calculates the mean and the variance of the qudrature operators
        for a single qumode state.
        The quadrature operators :math:`x` and :math:`p` for a mode :math:`i`
        can be calculated using the creation and annihilation operators as follows:

        .. math::
            x_i &= \sqrt{\frac{\hbar}{2}} (a_i + a_i^\dagger) \\
                p_i &= -i \sqrt{\frac{\hbar}{2}} (a_i - a_i^\dagger).

        Let :math:`\phi \in [ 0, 2 \pi )`, we can rotate the quadratures using the
        following transformation:

        .. math::
            Q_{i, \phi} = \cos\phi~x_i + \sin\phi~p_i.

        The expectation value :math:`\langle Q_{i, \phi}\rangle` can be calculated as:

        .. math::
            \operatorname{Tr}(\rho_i Q_{i, \phi}),

        where :math:`\rho_i` is the reduced density matrix of the mode :math:`i` and
        :math:`Q_{i, \phi}` is the rotated quadrature operator for a single mode and
        the variance is calculated as:

        .. math::
            \operatorname{\textit{Var}(Q_{i,\phi})} = \langle Q_{i, \phi}^{2}\rangle
                - \langle Q_{i, \phi}\rangle^{2}.

        Args:
            phi (float): The rotation angle. By default it is `0` which means that
                the mean of the position operator is being calculated. For :math:`\phi=
                \frac{\pi}{2}` the mean of the momentum operator is being calculated.
            modes (tuple[int]): The correspoding mode at which the mean of the
                quadratures are being calculated.
        Returns:
            (float, float): A tuple that contains the expectation value and the
                varianceof of the quadrature operator respectively.
        """
        reduced_dm = self.reduced(modes=modes).density_matrix
        annih = np.diag(np.sqrt(np.arange(1, self._config.cutoff)), 1)
        create = annih.T
        position = (create + annih) * np.sqrt(self._config.hbar / 2)
        momentum = -1j * (annih - create) * np.sqrt(self._config.hbar / 2)

        if phi != 0:
            rotated_quadratures = position * np.cos(phi) + momentum * np.sin(phi)
        else:
            rotated_quadratures = position

        expectation = np.trace(np.dot(reduced_dm, rotated_quadratures)).real
        variance = (
            np.trace(
                np.dot(reduced_dm, np.dot(rotated_quadratures, rotated_quadratures))
            ).real
            - expectation**2
        )
        return expectation, variance

    def get_purity(self):
        np = self._connector.np
        density_matrix = self.density_matrix
        return np.real(np.einsum("ij,ji", density_matrix, density_matrix))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from scipy.optimize import root_scalar

from piquasso._math.symplectic import xp_symplectic_form
from piquasso._math.transformations import xpxp_to_xxpp_indices

from piquasso.api.exceptions import InvalidParameter
from piquasso.api.connector import BaseConnector


def takagi(matrix, connector, atol=1e-12):
    """Takagi factorization of complex symmetric matrices.

    Note:

        The singular values have to be rounded due to floating point errors.

        The result is not unique in a sense that different result could be obtained
        by different ordering of the singular values.

    References:
    - https://journals.aps.org/pra/abstract/10.1103/PhysRevA.94.062109
    """

    np = connector.np

    V, singular_values, W_adjoint = connector.svd(matrix)

    W = np.conj(W_adjoint).T

    singular_value_multiplicity_indices = []
    singular_value_multiplicity_values = []

    for index, value in enumerate(singular_values):
        matches = np.where(
            np.isclose(value, np.array(singular_value_multiplicity_values), atol=atol)
        )[0]

        if len(matches) == 0:
            singular_value_multiplicity_values.append(value)
            singular_value_multiplicity_indices.append([index])
        else:
            singular_value_multiplicity_indices[matches[0]].append(index)

    diagonal_blocks_for_Q = []

    for indices in singular_value_multiplicity_indices:
        Z = V[:, indices].transpose() @ W[:, indices]

        diagonal_blocks_for_Q.append(connector.sqrtm(Z))

    Q = connector.block_diag(*diagonal_blocks_for_Q)

    return singular_values, V @ np.conj(Q)


def _rotation_to_positive_above_diagonals(block_diagonal_matrix, connector):
    """
    The block diagonal matrix returned by the Schur decomposition in the Williamson
    decomposition needs to be rotated.

    Not doing this we'd still get a valid Williamson decompostion with valid symplectic
    and diagonal matrices, but the symplectic matrix would have complex elements and the
    diagonal matrix would have negative values.
    """

    np = connector.np

    d = len(block_diagonal_matrix) // 2
    identity = np.identity(2)
    rotation = np.rot90(identity)

    return connector.block_diag(
        *[
            (
                identity
                if block_diagonal_matrix[2 * index, 2 * index + 1] > 0
                else rotation
            )
            for index in range(d)
        ]
    )


def williamson(matrix: np.ndarray, connector: BaseConnector) -> tuple:
    r"""
    Decomposes a positive definite matrix with Williamson decomposition, i.e. a
    positive definite :math:`M` is decomposed to

    .. math::

        M = S D S^T,

    where :math:`S \in \operatorname{Sp}(\mathbb{R}, 2d)` is a real symplectic matrix,
    and :math:`D` is a diagonal matrix containing positive values in the diagonal.

    The algorithm works as follows: without loss of generality, one can write the
    symplectic matrix in the form of

    .. math::

        S = M^{1 / 2} K D^{- 1 / 2}

    with :math:`K \in O(2d)`, since then

    .. math::

        M = S D S^T

    by construction. Now we need to find such :math:`K` that the value of :math:`S` is
    symplectic.

    .. math::

        S^T \Omega S = \Omega
            \rightleftarrow
            M^{- 1 / 2} J M^{- 1 / 2} = K D^{- 1 / 2} J D^{- 1 / 2} K^T,

    where

    .. math::

        D^{- 1 / 2} J D^{- 1 / 2}
            =
            \begin{bmatrix}
                0            & \hat{D}^{-1} \\
                \hat{D}^{-1} & 0            \\
            \end{bmatrix}

    is an antisymmetric matrix. We also know that :math:`M^{- 1 / 2} J M^{- 1 / 2}` is
    also antisymmetric. We just need to deduce the orthogonal transformation :math:`K`
    to acquire the symplectic matrix :math:`S`.

    We can use a (real) Schur decomposition to block-diagonalize
    :math:`M^{- 1 / 2} J M^{- 1 / 2}`. Note, that we also rotate the block to have the
    positive values in the above the diagonal to acquire real-valued symplectic matrices
    in the Williamson decomposition. Finally, we can acquire
    :math:`D^{- 1 / 2} J D^{- 1 / 2}` with a simple basis change.

    References:
        - https://math.stackexchange.com/a/1244793

    Args:
        matrix (numpy.ndarray): The matrix to decompose.

    Returns
        tuple: Tuple of the symplectic and diagonal matrices, in this order.
    """
    np = connector.np

    d = len(matrix) // 2

    omega = xp_symplectic_form(d)

    root_matrix = connector.sqrtm(matrix).real
    inverse_root_matrix = np.linalg.inv(root_matrix)

    block_diagonal_part, orthogonal_part = connector.schur(
        inverse_root_matrix @ omega @ inverse_root_matrix,
        output="real",
    )

    indices = xpxp_to_xxpp_indices(d)
    basis_change = _rotation_to_positive_above_diagonals(
        block_diagonal_part, connector
    )[:, indices]
    ordered_block_diagonal = basis_change.T @ block_diagonal_part @ basis_change

    inverse_diagonal_matrix = connector.block_diag(
        *(ordered_block_diagonal[:d, d:],) * 2
    )

    root_inverse_diagonal_matrix = np.diag(np.sqrt(np.diag(inverse_diagonal_matrix)))

    symplectic = (
        root_matrix @ orthogonal_part @ basis_change @ root_inverse_diagonal_matrix
    )

    diagonal_matrix = np.diag(1 / np.diag(inverse_diagonal_matrix))

    return symplectic, diagonal_matrix


def decompose_adjacency_matrix_into_circuit(
    adjacency_matrix, mean_photon_number, connector
):
    singular_values, unitary = takagi(adjacency_matrix, connector)

    scaling = _get_scaling(singular_values, mean_photon_number, adjacency_matrix)

    squeezing_parameters = np.arctanh(scaling * singular_values)

    return squeezing_parameters, unitary


def _get_scaling(
    singular_values: np.ndarray, mean_photon_number: float, adjacency_matrix: np.ndarray
) -> float:
    r"""
    For a squeezed state :math:`rho` the mean photon number is calculated by

    .. math::
        \langle n \rangle_\rho = \sum_{i = 0}^d \mathrm{sinh}(r_i)^2

    where :math:`r_i = \mathrm{arctan}(s_i)`, where :math:`s_i` are the singular
    values of the adjacency matrix.
    """

    def mean_photon_number_equation(scaling: float) -> float:
        return (
            sum(
                (scaling * singular_value) ** 2 / (1 - (scaling * singular_value) ** 2)
                for singular_value in singular_values
            )
            / len(singular_values)
            - mean_photon_number
        )

    def mean_photon_number_gradient(scaling: float) -> float:
        return (2.0 / scaling) * np.sum(
            (singular_values * scaling / (1 - (singular_values * scaling) ** 2)) ** 2
        )

    lower_bound = 0.0

    tolerance = 1e-10  # Needed to avoid zero division.

    upper_bound = 1.0 / (max(singular_values) + tolerance)

    result = root_scalar(
        mean_photon_number_equation,
        fprime=mean_photon_number_gradient,
        x0=(lower_bound - upper_bound) / 2.0,
        bracket=(lower_bound, upper_bound),
    )

    if not result.converged:
        raise InvalidParameter(
            f"No scaling found for adjacency matrix: {adjacency_matrix}."
        )

    return result.root


def euler(symplectic, connector):
    np = connector.np
    d = len(symplectic) // 2

    U_orig, R = connector.polar(symplectic, side="left")

    K = np.diag(np.array([1.0] * d + [-1.0] * d))

    H_active = 1j * K @ connector.logm(R)

    Z = 1j * H_active[:d, d:]

    D, U = takagi(Z, connector)

    return U, D, np.conj(U).T @ U_orig[:d, :d]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import jax
import jax.numpy as jnp


def factorial(n):
    return jnp.where(n < 0, 0, jax.lax.exp(jax.lax.lgamma(n + 1)))


# The following implementation of `eig` is copied with copyright from
# https://github.com/facebookresearch/fmmax/blob/8c7491a23e902644614c1761b6fbefa472786567/src/fmmax/utils.py#L101
# following discussions on https://github.com/jax-ml/jax/issues/2748.
# Minor modifications were applied.


# MIT License
#
# Copyright (c) 2023 Meta Research
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


EPS_EIG = 1e-6


@jax.custom_vjp
def eig(matrix: jnp.ndarray, eps: float = EPS_EIG) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Wraps `jnp.linalg.eig` in a jit-compatible, differentiable manner.

    The custom vjp allows gradients with resepct to the eigenvectors, unlike the
    standard jax implementation of `eig`. We use an expression for the gradient
    given in [2019 Boeddeker] along with a regularization scheme used in [2021
    Colburn]. The method effectively applies a Lorentzian broadening to a term
    containing the inverse difference of eigenvalues.

    [2019 Boeddeker] https://arxiv.org/abs/1701.00392
    [2021 Coluburn] https://www.nature.com/articles/s42005-021-00568-6

    Args:
        matrix: The matrix for which eigenvalues and eigenvectors are sought.
        eps: Parameter which determines the degree of broadening.

    Returns:
        The eigenvalues and eigenvectors.
    """
    del eps
    return _eig_host(matrix)


def _eig_host(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Wraps jnp.linalg.eig so that it can be jit-ed on a machine with GPUs."""
    eigenvalues_shape = jax.ShapeDtypeStruct(matrix.shape[:-1], matrix.dtype)
    eigenvectors_shape = jax.ShapeDtypeStruct(matrix.shape, matrix.dtype)

    def _eig_cpu(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        # We force this computation to be performed on the cpu by jit-ing and
        # explicitly specifying the device.
        with jax.default_device(jax.devices("cpu")[0]):
            return jax.jit(jnp.linalg.eig)(matrix)

    return jax.pure_callback(
        _eig_cpu,
        (eigenvalues_shape, eigenvectors_shape),
        matrix,
    )


def _eig_fwd(
    matrix: jnp.ndarray,
    eps: float,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray, float]]:
    """Implements the forward calculation for `eig`."""
    eigenvalues, eigenvectors = _eig_host(matrix)
    return (eigenvalues, eigenvectors), (eigenvalues, eigenvectors, eps)


def _eig_bwd(
    res: Tuple[jnp.ndarray, jnp.ndarray, float],
    grads: Tuple[jnp.ndarray, jnp.ndarray],
) -> Tuple[jnp.ndarray, None]:
    """Implements the backward calculation for `eig`."""
    eigenvalues, eigenvectors, eps = res
    grad_eigenvalues, grad_eigenvectors = grads

    # Compute the F-matrix, from equation 5 of [2021 Colburn]. This applies a
    # Lorentzian broadening to the matrix `f = 1 / (eigenvalues[i] - eigenvalues[j])`.
    eigenvalues_i = eigenvalues[..., jnp.newaxis, :]
    eigenvalues_j = eigenvalues[..., :, jnp.newaxis]
    f_broadened = (eigenvalues_i - eigenvalues_j) / (
        (eigenvalues_i - eigenvalues_j) ** 2 + eps
    )

    # Manually set the diagonal elements to zero, as we do not use broadening here.
    i = jnp.arange(f_broadened.shape[-1])
    f_broadened = f_broadened.at[..., i, i].set(0)

    # By jax convention, gradients are with respect to the complex parameters, not with
    # respect to their conjugates. Take the conjugates.
    grad_eigenvalues_conj = jnp.conj(grad_eigenvalues)
    grad_eigenvectors_conj = jnp.conj(grad_eigenvectors)

    eigenvectors_H = jnp.conj(eigenvectors).T
    dim = eigenvalues.shape[-1]
    eye_mask = jnp.eye(dim, dtype=bool)
    eye_mask = eye_mask.reshape((1,) * (eigenvalues.ndim - 1) + (dim, dim))

    # Then, the gradient is found by equation 4.77 of [2019 Boeddeker].
    rhs = (
        jnp.diag(grad_eigenvalues_conj)
        + jnp.conj(f_broadened) * (eigenvectors_H @ grad_eigenvectors_conj)
        - jnp.conj(f_broadened)
        * (eigenvectors_H @ eigenvectors)
        @ jnp.where(eye_mask, jnp.real(eigenvectors_H @ grad_eigenvectors_conj), 0.0)
    ) @ eigenvectors_H
    grad_matrix = jnp.linalg.solve(eigenvectors_H, rhs)

    # Take the conjugate of the gradient, reverting to the jax convention
    # where gradients are with respect to complex parameters.
    grad_matrix = jnp.conj(grad_matrix)

    # Return `grad_matrix`, and `None` for the gradient with respect to `eps`.
    return grad_matrix, None


eig.defvjp(_eig_fwd, _eig_bwd)
